"""
gpx_dialog.py

This is the pop-up dialog for the GPX Viewer function.

Created on 10 Jan 2023

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

# pylint: disable=unused-argument

import logging
import traceback
from statistics import mean, median
from tkinter import (
    ALL,
    NE,
    NW,
    SE,
    SW,
    Button,
    Canvas,
    E,
    Frame,
    IntVar,
    Label,
    N,
    S,
    Spinbox,
    StringVar,
    W,
)
from xml.dom import minidom
from xml.parsers import expat

from pynmeagps import haversine, planar

from pygpsclient.globals import (
    AXISCOL,
    BGCOL,
    CUSTOM,
    ERRCOL,
    FGCOL,
    GRIDCOL,
    HOME,
    INFOCOL,
    M2FT,
    M2KM,
    M2MIL,
    M2NMIL,
    MPS2KNT,
    MPS2KPH,
    MPS2MPH,
    READONLY,
    ROUTE,
    TRACK,
    UI,
    UIK,
    UMK,
    WAYPOINT,
    Area,
    AreaXY,
    Point,
    TrackPoint,
)
from pygpsclient.helpers import (
    data2xy,
    fontheight,
    get_grid,
    isot2dt,
    time2str,
)
from pygpsclient.map_canvas import HYB, MAP, SAT, MapCanvas
from pygpsclient.strings import (
    DLGGPXERROR,
    DLGGPXLOAD,
    DLGGPXLOADED,
    DLGGPXNOMINAL,
    DLGGPXNULL,
    DLGGPXOPEN,
    DLGGPXWAIT,
    DLGTGPX,
    NA,
)
from pygpsclient.toplevel_dialog import ToplevelDialog

# profile chart parameters:
AXIS_XL = 35  # x axis left offset
AXIS_XR = 35  # x axis right offset
AXIS_Y = 15  # y axis bottom offset
ELEAX_COL = "green4"  # color of elevation plot axis
ELE_COL = "palegreen3"  # color of elevation plot
SPD_COL = "coral"  # color of speed plot
TRK_COL = "magenta"  # color of track
MD_LINES = 3  # number of lines of metadata
MINDIM = (400, 500)

GPXTYPES = {TRACK: "trkpt", WAYPOINT: "wpt", ROUTE: "rtept"}
AXISTAG = "axt"
AXISLBLTAG = "axl"
ERRCOL = "coral"
LBLCOL = "white"
CONTRASTCOL = "black"
CHARTMINY = 0
CHARTMAXY = 100
CHARTSCALE = 1
MAXCHANS = 2
CHANELE = 0
CHANSPD = 1
RESFONT = 28  # font size relative to widget size
MINFONT = 8  # minimum font size
PLOTWID = 1
PLOTCOLS = ("yellow", "cyan", "magenta", "deepskyblue")
GRIDMINCOL = "grey30"
LBLGRID = 5
GRIDSTEPS = get_grid(5)
XLBLSTEPS = get_grid(LBLGRID)
YLBLSTEPS = get_grid(3)
YLBLRANGE = {
    0: (50, 100, 250, 500, 1000, 2500, 5000, 10000, 25000),
    1: (5, 10, 25, 50, 100, 200, 500, 1000),
}


class GPXViewerDialog(ToplevelDialog):
    """GPXViewerDialog class."""

    def __init__(self, app, *args, **kwargs):
        """Constructor."""

        self.__app = app
        self.logger = logging.getLogger(__name__)
        # self.__master = self.__app.appmaster  # link to root Tk window
        super().__init__(app, DLGTGPX, MINDIM)
        self._mapzoom = IntVar()
        self._maptype = StringVar()
        self._gpxtype = StringVar()
        zoom = int(kwargs.get("zoom", 12))
        self._mapzoom.set(zoom)
        self._info = []
        for _ in range(MD_LINES):
            self._info.append(StringVar())
        self._mtt = None
        self._mtz = None
        self._mtg = None
        self._mapimg = None
        self._track = None
        self._gpxfile = None
        self._bounds = None
        self._center = None
        self._initdir = HOME
        self._no_time = False
        self._no_ele = False

        # elevation/speed profile parameters
        self._font = self.__app.font_vsm
        self._fonth = fontheight(self._font)
        self._num_chans = 2
        self._xoff = 20  # chart X offset for labels
        self._yoff = 20  # chart Y offset for labels
        self._plotcols = PLOTCOLS
        self._mintim = 1e20
        self._maxtim = 0
        self._rng = 0
        self._elapsed = 0
        self._dist = 0
        self._minele = self._minspd = 1e20
        self._maxele = self._maxspd = -1e20

        self._body()
        self._do_layout()
        self._reset()
        self._attach_events()
        self._init_profile()
        self._finalise()

    def _body(self):
        """
        Create widgets.
        """

        self._frm_body = Frame(self.container, borderwidth=2, relief="groove")
        self._frm_map = Frame(self._frm_body, borderwidth=2, relief="groove", bg=BGCOL)
        self._frm_profile = Frame(
            self._frm_body, borderwidth=2, relief="groove", bg=BGCOL
        )
        self._frm_info = Frame(self._frm_body, borderwidth=2, relief="groove", bg=BGCOL)
        self._frm_controls = Frame(self._frm_body, borderwidth=2, relief="groove")
        self._can_mapview = MapCanvas(
            self.__app,
            self._frm_map,
            height=self.height * 0.75,
            width=self.width,
            bg=BGCOL,
        )
        self._can_profile = Canvas(
            self._frm_profile,
            height=self.height * 0.25,
            width=self.width,
            bg=BGCOL,
        )
        self._lbl_info = []
        for i in range(MD_LINES):
            fg = PLOTCOLS[i - 1] if i else FGCOL
            self._lbl_info.append(
                Label(
                    self._frm_info,
                    textvariable=self._info[i],
                    anchor=W,
                    bg=BGCOL,
                    fg=fg,
                )
            )
        self._btn_load = Button(
            self._frm_controls,
            image=self.img_load,
            width=40,
            command=self._on_load,
        )
        self._lbl_maptype = Label(self._frm_controls, text="Map Type")
        self._spn_maptype = Spinbox(
            self._frm_controls,
            values=(HYB, SAT, MAP, CUSTOM),
            width=7,
            wrap=True,
            textvariable=self._maptype,
            state=READONLY,
        )
        self._spn_gpxtype = Spinbox(
            self._frm_controls,
            values=list(GPXTYPES.keys()),
            width=6,
            wrap=True,
            textvariable=self._gpxtype,
            state=READONLY,
        )
        self._lbl_zoom = Label(self._frm_controls, text="Zoom")
        self._spn_zoom = Spinbox(
            self._frm_controls,
            from_=1,
            to=20,
            width=5,
            wrap=False,
            textvariable=self._mapzoom,
            state=READONLY,
        )
        self._btn_redraw = Button(
            self._frm_controls,
            image=self.img_redraw,
            width=40,
            command=self._on_redraw,
        )

    def _do_layout(self):
        """
        Arrange widgets.
        """

        self._frm_body.grid(column=0, row=0, sticky=(N, S, E, W))
        self._frm_map.grid(column=0, row=0, sticky=(N, S, E, W))
        self._frm_profile.grid(column=0, row=1, sticky=(W, E))
        self._frm_info.grid(column=0, row=2, sticky=(W, E))
        self._frm_controls.grid(column=0, row=3, columnspan=7, sticky=(W, E))
        self._can_mapview.grid(column=0, row=0, sticky=(N, S, E, W))
        self._can_profile.grid(column=0, row=0, sticky=(N, S, E, W))
        for i in range(MD_LINES):
            self._lbl_info[i].grid(column=0, row=i, padx=1, pady=1, sticky=(W, E))
        self._btn_load.grid(column=0, row=1, padx=3, pady=3)
        self._lbl_maptype.grid(
            column=1,
            row=1,
            padx=3,
            pady=3,
        )
        self._spn_maptype.grid(
            column=2,
            row=1,
            padx=3,
            pady=3,
        )
        self._spn_gpxtype.grid(
            column=3,
            row=1,
            padx=3,
            pady=3,
        )
        self._lbl_zoom.grid(
            column=4,
            row=1,
            padx=3,
            pady=3,
        )
        self._spn_zoom.grid(
            column=5,
            row=1,
            padx=3,
            pady=3,
        )
        self._btn_redraw.grid(
            column=6,
            row=1,
            padx=3,
            pady=3,
        )

        # set column and row weights
        # NB!!! these govern the 'pack' behaviour of the frames on resize
        self.container.grid_columnconfigure(0, weight=1)
        self.container.grid_rowconfigure(0, weight=1)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self._frm_body.grid_columnconfigure(0, weight=1)
        self._frm_body.grid_rowconfigure(0, weight=20)  # map
        self._frm_body.grid_rowconfigure(1, weight=1)  # profile
        self._frm_body.grid_rowconfigure(2, weight=0)  # info
        self._frm_map.grid_columnconfigure(0, weight=1)
        self._frm_map.grid_rowconfigure(0, weight=1)
        self._frm_profile.grid_columnconfigure(0, weight=1)
        self._frm_profile.grid_rowconfigure(0, weight=1)

    def _attach_events(self):
        """
        Bind events to window.
        """

        self._mtt = self._maptype.trace_add("write", self._on_maptype)
        self._mtz = self._mapzoom.trace_add("write", self._on_mapzoom)
        self._mtg = self._gpxtype.trace_add("write", self._on_gpxtype)

    def _detach_events(self):
        """
        Unbind events to window.
        """

        if self._mtt is not None:
            self._maptype.trace_remove("write", self._mtt)
        if self._mtz is not None:
            self._mapzoom.trace_remove("write", self._mtz)
        if self._mtg is not None:
            self._gpxtype.trace_remove("write", self._mtg)

    def _reset(self):
        """
        Reset application.
        """

        self._can_mapview.delete(ALL)
        self._can_profile.delete(ALL)
        self._maptype.set(self.__app.configuration.get("gpxmaptype_s"))
        self._mapzoom.set(self.__app.configuration.get("gpxmapzoom_n"))
        self._gpxtype.set(self.__app.configuration.get("gpxtype_s"))
        for i in range(MD_LINES):
            self._info[i].set("")
        self._can_mapview.draw_msg(DLGGPXOPEN, INFOCOL)

    def _on_maptype(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Map type has changed.
        """

        self.__app.configuration.set("gpxmaptype_s", self._maptype.get())
        self._on_redraw()

    def _on_mapzoom(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Map zoom has changed.
        """

        self.__app.configuration.set("gpxmapzoom_n", self._mapzoom.get())
        self._on_redraw()

    def _on_gpxtype(self, var, index, mode):  # pylint: disable=unused-argument
        """
        GPX type has changed.
        """

        self.__app.configuration.set("gpxtype_s", self._gpxtype.get())

    def _on_redraw(self, *args, **kwargs):
        """
        Handle redraw button press.
        """

        self.set_status(DLGGPXWAIT, INFOCOL)
        self._detach_events()
        self._reset()
        self._spn_zoom.config(highlightbackground="gray90", highlightthickness=3)
        self._parse_gpx()
        self._draw_map()
        self._draw_profile()
        self._draw_metadata()
        self._attach_events()

    def _open_gpxfile(self) -> str:
        """
        Open gpx file.
        """

        return self.__app.file_handler.open_file(
            "gpx",
            (("gpx files", "*.gpx"), ("all files", "*.*")),
        )

    def _on_load(self):
        """
        Load gpx track from file.
        """

        self._gpxfile = self._open_gpxfile()
        if self._gpxfile is None:  # user cancelled
            return
        self.set_status(DLGGPXLOAD, INFOCOL)
        self._parse_gpx()

    def _parse_gpx(self):

        if self._gpxfile is None:
            return
        ptyp = GPXTYPES[self._gpxtype.get()]
        with open(self._gpxfile, "r", encoding="utf-8") as gpx:
            try:
                parser = minidom.parse(gpx)
                trkpts = parser.getElementsByTagName(f"{ptyp}")
                self._process_track(trkpts, ptyp)
            except (TypeError, AttributeError, expat.ExpatError) as err:
                self.set_status(f"{DLGGPXERROR}\n{repr(err)}", ERRCOL)
                self.logger.error(traceback.format_exc())

    def _process_track(self, trkpts: list, ptyp: str):
        """
        Process trackpoint data.

        :param list trk: list of trackpoints
        :param str ptyp: element type
        """

        self._rng = len(trkpts)
        self._no_time = False
        self._no_ele = False
        if self._rng == 0:
            self.set_status(DLGGPXNULL.format(ptyp), ERRCOL)
            return

        minlat = minlon = 400
        maxlat = maxlon = -400
        track = []
        start = end = 0
        self._dist = 0
        lat1 = lon1 = 0
        tm = tim1 = spd = spd1 = 0
        self._minele = self._minspd = 1e10
        self._maxele = self._maxspd = -1e20
        for i, trkpt in enumerate(trkpts):
            lat = float(trkpt.attributes["lat"].value)
            lon = float(trkpt.attributes["lon"].value)
            # establish bounding box
            minlat = min(minlat, lat)
            minlon = min(minlon, lon)
            maxlat = max(maxlat, lat)
            maxlon = max(maxlon, lon)
            try:
                tim = isot2dt(trkpt.getElementsByTagName("time")[0].firstChild.data)
            except IndexError:  # time element does not exist
                self._no_time = True
                tim = tm + i  # use synthetic timestamp if gpx has no time element
            if i == 0:
                lat1, lon1, tim1 = lat, lon, tim
                start = tim
            else:
                leg = planar(lat1, lon1, lat, lon)  # m
                if leg > 1000:
                    leg = haversine(lat1, lon1, lat, lon) / 1000  # m
                self._dist += leg
                if tim > tim1:
                    spd = leg / (tim - tim1)  # m/s
                else:
                    spd = spd1
                spd1 = spd
                self._maxspd = max(spd, self._maxspd)
                self._minspd = min(spd, self._minspd)
                lat1, lon1, tim1 = lat, lon, tim
                end = tim
            try:
                ele = float(trkpt.getElementsByTagName("ele")[0].firstChild.data)
                self._maxele = max(ele, self._maxele)
                self._minele = min(ele, self._minele)
            except IndexError:  # 'ele' element does not exist
                self._no_ele = True
                ele = 0.0

            track.append(TrackPoint(lat, lon, tim, ele, spd))

        self._bounds = Area(minlat, minlon, maxlat, maxlon)
        self._center = Point((maxlat + minlat) / 2, (maxlon + minlon) / 2)
        self._elapsed = end - start
        self._mintim = start
        self._maxtim = end
        self._track = track

        self._draw_map()
        self._draw_profile()
        self._draw_metadata()
        self.set_status(DLGGPXLOADED, INFOCOL)

    def _get_units(self) -> tuple:
        """
        Get speed and elevation units and conversions.
        Default is metric - meters and meters per second.

        :return: tuple of (dst_u, dst_c, ele_u, ele_C, spd_u, spd_c)
        :rtype: tuple
        """

        units = self.__app.configuration.get("units_s")

        if units == UI:
            dst_u = "miles"
            dst_c = M2MIL
            ele_u = "ft"
            ele_c = M2FT
            spd_u = "mph"
            spd_c = MPS2MPH
        elif units == UIK:
            dst_u = "naut miles"
            dst_c = M2NMIL
            ele_u = "ft"
            ele_c = M2FT
            spd_u = "knt"
            spd_c = MPS2KNT
        elif units == UMK:
            dst_u = "km"
            dst_c = M2KM
            ele_u = "m"
            ele_c = 1
            spd_u = "kph"
            spd_c = MPS2KPH
        else:  # UMM
            dst_u = "m"
            dst_c = 1
            ele_u = "m"
            ele_c = 1
            spd_u = "m/s"
            spd_c = 1

        return dst_u, dst_c, ele_u, ele_c, spd_u, spd_c

    def _draw_map(self):
        """
        Draw map on canvas.
        """

        if self._track is None:
            return
        location = self._center
        maptype = self._maptype.get()
        zoom = self._mapzoom.get()
        bounds = self._can_mapview.zoom_bounds(
            self.height, self.width, location, zoom, maptype
        )
        points = [Point(pnt.lat, pnt.lon) for pnt in self._track]
        self._can_mapview.draw_map(
            maptype,
            location=location,
            track=points,
            bounds=bounds,
            zoom=zoom,
            marker=self._can_mapview.marker,
        )

        if self._can_mapview.zoommin:
            self._spn_zoom.config(highlightbackground=ERRCOL, highlightthickness=3)
        else:
            self._spn_zoom.config(highlightbackground="gray90", highlightthickness=3)

    def _init_profile(self):
        """
        Initialise elevation/speed profile.
        """

        self.update_idletasks()  # Make sure we know about any resizing
        w, h = self._can_profile.winfo_width(), self._can_profile.winfo_height()
        self._xoff = self._fonth * self._num_chans / 2 + 3  # chart X offset for labels
        self._yoff = self._fonth + 3  # chart Y offset for labels
        self._can_profile.delete(ALL)

        # draw grid
        for i, p in enumerate(GRIDSTEPS):
            y = (h - self._yoff) * p
            col = AXISCOL if p in (0, 1.0) else GRIDMINCOL if i % LBLGRID else GRIDCOL
            self._can_profile.create_line(
                self._xoff, y, w - self._xoff, y, fill=col, tags=AXISTAG
            )
            x = self._xoff + (w - self._xoff * 2) * p
            self._can_profile.create_line(
                x, 0, x, h - self._yoff, fill=col, tags=AXISTAG
            )

    def _draw_profile(self):
        """
        Update elevation/speed profile with data.
        """

        if self._track is None:
            return
        self._init_profile()

        w, h = self._can_profile.winfo_width(), self._can_profile.winfo_height()
        _, _, ele_u, ele_c, spd_u, spd_c = self._get_units()

        # set default ranges for all channels
        minval = [CHARTMINY] * self._num_chans
        maxval = [CHARTMAXY] * self._num_chans
        scale = [CHARTSCALE] * self._num_chans
        label = [""] * self._num_chans
        scale[CHANELE] = ele_c
        label[CHANELE] = f"hmsl ({ele_u})"
        scale[CHANSPD] = spd_c
        label[CHANSPD] = f"speed ({spd_u})"

        # get X axis (time) range for all channels and draw labels
        mintim, maxtim = self._mintim, self._maxtim
        bounds = AreaXY(mintim, CHARTMINY, maxtim, CHARTMAXY)
        self._can_profile.delete(AXISLBLTAG)
        self._draw_xaxis_labels(w, h, bounds, mintim, maxtim)

        # plot each channel's data points
        for chn in range(self._num_chans):

            chncol = self._plotcols[chn]
            minval[chn] = 0
            # autorange max y values
            if chn == CHANELE:
                for i in YLBLRANGE[chn]:
                    if self._maxele * ele_c < i:
                        maxval[chn] = i
                        break
            elif chn == CHANSPD:
                for i in YLBLRANGE[chn]:
                    if self._maxspd * spd_c < i:
                        maxval[chn] = i
                        break

            bounds = AreaXY(mintim, minval[chn], maxtim, maxval[chn])

            # draw Y axis (data value) labels for this channel
            self._draw_yaxis_labels(
                w, h, bounds, minval[chn], maxval[chn], chn, label[chn]
            )

            # plot each track element
            inr = False
            for pnt in self._track:
                tim = pnt.tim
                try:
                    val = pnt.ele if chn == CHANELE else pnt.spd
                except KeyError:
                    val = None

                if val is None:
                    continue

                if scale[chn] != 1:
                    val *= scale[chn]  # scale data

                # convert datapoint to canvas x,y coordinates
                x, y = data2xy(
                    w - self._xoff * 2,
                    h - self._yoff,
                    bounds,
                    tim,
                    val,
                    self._xoff,
                )
                if x <= self._xoff:
                    inr = False
                # plot line
                if inr:
                    x2, y2 = x, y
                    self._can_profile.create_line(
                        x1,
                        y1,
                        x2,
                        y2,
                        fill=chncol,
                        width=PLOTWID,
                        tags=f"plot_{chn:1d}",
                    )
                    x1, y1 = x2, y2
                else:
                    x1, y1 = max(x, self._xoff), y
                inr = True

        if self._no_time:
            self._timelegend(w)

    def _draw_xaxis_labels(
        self, w: int, h: int, bounds: AreaXY, mintim: float, maxtim: float
    ):
        """
        Draw X axis (time) labels.

        :param int w: canvas width
        :param int h: canvas height
        :param AreaXY bounds: data bounds
        :param float mintim: minimum time
        :param float maxtim: maximum time
        """
        # pylint: disable=too-many-arguments, too-many-positional-arguments

        for g in XLBLSTEPS:
            xval = mintim + (maxtim - mintim) * g
            x, _ = data2xy(w - self._xoff * 2, h - self._yoff, bounds, xval, 0)
            if g == 0:
                anc = NW
            elif g == 1:
                anc = NE
            else:
                anc = N
            self._can_profile.create_text(
                x + self._xoff,
                h - self._yoff,
                text=time2str(xval),
                anchor=anc,
                fill=AXISCOL,
                font=self._font,
                tags=AXISLBLTAG,
            )

    def _draw_yaxis_labels(
        self,
        w: int,
        h: int,
        bounds: AreaXY,
        minval: float,
        maxval: float,
        chn: int,
        lbl: str,
    ):
        """
        Draw Y axis (data value) labels for this channel.

        :param int w: canvas width
        :param int h: canvas height
        :param AreaXY bounds: data bounds
        :param float minval: minimum val for chn
        :param float maxval: maximum val for chn
        :param int chn: channel
        """
        # pylint: disable=too-many-arguments, too-many-positional-arguments

        col = self._plotcols[chn]
        yo = 2  # avoid edges
        # y axis labels alternate left and right
        if chn % 2:  # odd channels
            x = w - yo - self._fonth * ((chn - 1) / 2)
        else:  # even channels
            x = yo + self._fonth * (chn / 2)

        for g in YLBLSTEPS:
            yval = minval + (maxval - minval) * g
            _, y = data2xy(
                w - self._xoff * MAXCHANS / 2, h - self._yoff, bounds, 0, yval
            )
            if g == 0:
                anc = SW if chn % 2 else NW
            elif g == 1:
                y += yo * 2  # avoid edges
                anc = SE if chn % 2 else NE
            else:
                anc = S if chn % 2 else N
            self._can_profile.create_text(
                x,
                y,
                text=int(yval),
                fill=col,
                font=self._font,
                angle=90,
                anchor=anc,
                tags=AXISLBLTAG,
            )

        # legend
        self._can_profile.create_text(
            w - yo * 2 - self._fonth if chn % 2 else yo * 2 + self._fonth,
            2,
            text=lbl,
            fill=self._plotcols[chn],
            font=self._font,
            anchor=NE if chn % 2 else NW,
            tags=AXISLBLTAG,
        )

    def _timelegend(self, w: int):
        """
        Draw nominal time legend.
        """

        self._can_profile.create_text(
            w / 2,
            2,
            text=DLGGPXNOMINAL,
            fill=AXISCOL,
            font=self._font,
            anchor=N,
            tags=AXISLBLTAG,
        )

    def _draw_metadata(self) -> str:
        """
        Draw metadata as lines of text.
        """

        if self._rng == 0:
            return

        dst_u, dst_c, ele_u, ele_c, spd_u, spd_c = self._get_units()

        if self._elapsed > 3600:
            elapsed = self._elapsed / 3600
            elp_u = "hours"
        else:
            elapsed = self._elapsed
            elp_u = "seconds"
        self._info[0].set(
            (
                f"Track elements: {self._rng:,}; "
                f"Distance ({dst_u}): {self._dist*dst_c:,.2f}; "
                f"Elapsed ({elp_u}): {elapsed:,.2f}"
            )
        )

        if self._no_ele:
            ele = NA
        else:
            elelist = [i.ele * ele_c for i in self._track]
            ele_median = median(elelist)
            ele_mean = mean(elelist)
            ele_min = min(elelist)
            ele_max = max(elelist)
            ele = (
                f"({ele_u}) min: {ele_min:,.2f} "
                f"max: {ele_max:,.2f} "
                f"avg: {ele_mean:,.2f} med: {ele_median:,.2f} "
                f"dif: {(ele_max-ele_min):,.2f}"
            )
        self._info[1].set(f"Ele {ele}")

        if self._no_time:
            spd = NA
        else:
            spdlist = [i.spd * spd_c for i in self._track]
            spd_median = median(spdlist)
            spd_mean = mean(spdlist)
            spd_min = min(spdlist)
            spd_max = max(spdlist)
            spd = (
                f"({spd_u}) min: {spd_min:,.2f} "
                f"max: {spd_max:,.2f} "
                f"avg: {spd_mean:,.2f} med: {spd_median:,.2f}"
            )
        self._info[2].set(f"Spd {spd}")



"""
tty_preset_dialog.py

TTY Configuration frame for user-defined TTY (AT+) commands

Created on 7 May 2025

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import (
    HORIZONTAL,
    LEFT,
    VERTICAL,
    Button,
    Checkbutton,
    E,
    Entry,
    Frame,
    IntVar,
    Label,
    Listbox,
    N,
    S,
    Scrollbar,
    StringVar,
    W,
    ttk,
)

from pygpsclient.confirm_box import ConfirmBox
from pygpsclient.globals import (
    ASCII,
    BSR,
    CRLF,
    ERRCOL,
    INFOCOL,
    OKCOL,
    TTY_EVENT,
    TTYERR,
    TTYOK,
)
from pygpsclient.strings import (
    CONFIRM,
    DLGACTION,
    DLGACTIONCONFIRM,
    DLGTTTY,
)
from pygpsclient.toplevel_dialog import ToplevelDialog

CANCELLED = 0
CONFIRMED = 1
NOMINAL = 2
MINDIM = (610, 500)


class TTYPresetDialog(ToplevelDialog):
    """
    TTY Preset and User-defined configuration command dialog.
    """

    def __init__(self, app, **kwargs):  # pylint: disable=unused-argument
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app
        self.__master = self.__app.appmaster  # Reference to root class (Tk)

        super().__init__(app, DLGTTTY, MINDIM)
        self._confirm = False
        self._command = StringVar()
        self._crlf = IntVar()
        self._echo = IntVar()
        self._delay = IntVar()
        self._body()
        self._do_layout()
        self.reset()
        self._attach_events()
        self._finalise()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self._frm_body = Frame(self.container, borderwidth=2, relief="groove")
        self._lbl_command = Label(
            self._frm_body,
            text="Command",
        )
        self._ent_command = Entry(
            self._frm_body,
            textvariable=self._command,
            relief="sunken",
            width=50,
        )
        self._chk_crlf = Checkbutton(
            self._frm_body,
            text="CRLF",
            variable=self._crlf,
        )
        self._chk_echo = Checkbutton(
            self._frm_body,
            text="Echo",
            variable=self._echo,
        )
        self._chk_delay = Checkbutton(
            self._frm_body,
            text="Delay",
            variable=self._delay,
        )
        self._lbl_presets = Label(self._frm_body, text="Preset TTY Commands", anchor=W)
        self._lbx_preset = Listbox(
            self._frm_body,
            border=2,
            relief="sunken",
            height=25,
            width=55,
            justify=LEFT,
            exportselection=False,
        )
        self._scr_presetv = Scrollbar(self._frm_body, orient=VERTICAL)
        self._scr_preseth = Scrollbar(self._frm_body, orient=HORIZONTAL)
        self._lbx_preset.config(yscrollcommand=self._scr_presetv.set)
        self._lbx_preset.config(xscrollcommand=self._scr_preseth.set)
        self._scr_presetv.config(command=self._lbx_preset.yview)
        self._scr_preseth.config(command=self._lbx_preset.xview)
        self._lbl_send_command = Label(self._frm_body)
        self._btn_send_command = Button(
            self._frm_body,
            image=self.img_send,
            width=50,
            command=self._on_send_command,
        )

    def _do_layout(self):
        """
        Layout widgets.
        """

        self._frm_body.grid(
            column=0, row=0, padx=5, pady=5, ipadx=5, ipady=5, sticky=(N, S, E, W)
        )
        self._lbl_command.grid(column=0, row=0, padx=3, sticky=W)
        self._ent_command.grid(column=1, row=0, columnspan=3, padx=3, sticky=(W, E))
        self._chk_crlf.grid(column=0, row=1, padx=3, sticky=W)
        self._chk_echo.grid(column=1, row=1, padx=3, sticky=W)
        self._chk_delay.grid(column=2, row=1, padx=3, sticky=W)
        ttk.Separator(self._frm_body).grid(
            column=0, row=2, columnspan=4, padx=2, pady=2, sticky=(W, E)
        )
        self._lbl_presets.grid(column=0, row=3, columnspan=3, padx=3, sticky=(W, E))
        self._lbx_preset.grid(
            column=0,
            row=3,
            columnspan=3,
            rowspan=10,
            padx=3,
            pady=3,
            sticky=(N, S, W, E),
        )
        self._scr_presetv.grid(column=2, row=3, rowspan=21, sticky=(N, S, E))
        self._scr_preseth.grid(column=0, row=24, columnspan=3, sticky=(W, E))
        self._btn_send_command.grid(
            column=3, row=3, padx=3, ipadx=3, ipady=3, sticky=(N, E)
        )
        self._lbl_send_command.grid(
            column=3, row=4, padx=3, ipadx=3, ipady=3, sticky=(N, W, E)
        )
        self.container.grid_columnconfigure(0, weight=10)
        self.container.grid_rowconfigure(0, weight=10)
        self.grid_columnconfigure(0, weight=10)
        self.grid_rowconfigure(0, weight=10)
        colsp, _ = self._frm_body.grid_size()
        for col in range(colsp - 1):
            self._frm_body.grid_columnconfigure(col, weight=10)
        self._frm_body.grid_rowconfigure(3, weight=10)

    def _attach_events(self):
        """
        Bind listbox selection events.
        """

        self._command.trace_add("write", self._on_update_command)
        for setting in (self._crlf, self._echo):
            setting.trace_add("write", self._on_update_settings)
        self._lbx_preset.bind("<<ListboxSelect>>", self._on_select_preset)
        # self.bind("<Configure>", self._on_resize)

    def reset(self):
        """
        Reset panel - Load user-defined presets if there are any.
        """

        self._crlf.set(self.__app.configuration.get("ttycrlf_b"))
        self._echo.set(self.__app.configuration.get("ttyecho_b"))
        self._delay.set(self.__app.configuration.get("ttydelay_b"))
        idx = 0
        for tcmd in self.__app.configuration.get("ttypresets_l"):
            self._lbx_preset.insert(idx, tcmd)
            idx += 1

    def _on_update_command(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Command has been updated.
        """

        self._lbl_send_command.config(image=self.img_none)

    def _on_update_settings(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Settings have been updated.
        """

        self.__app.configuration.set("ttycrlf_b", self._crlf.get())
        self.__app.configuration.set("ttyecho_b", self._echo.get())
        self.__app.configuration.set("ttydelay_b", self._delay.get())

    def _on_select_preset(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Preset command has been selected.
        """

        try:
            self.set_status("", INFOCOL)
            idx = self._lbx_preset.curselection()
            preset = self._lbx_preset.get(idx).split(";", 1)
            self._confirm = CONFIRM in preset[0]
            self._command.set(preset[1])
        except IndexError:
            self.set_status("Invalid preset format", ERRCOL)

    def _on_send_command(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Preset command send button has been clicked.
        """

        if self._command.get() in ("", None):
            self.set_status("Enter or select command", ERRCOL)
            return

        try:
            if self._confirm:
                if ConfirmBox(self, DLGACTION, DLGACTIONCONFIRM).show():
                    self._parse_command(self._command.get())
                    status = CONFIRMED
                else:
                    status = CANCELLED
            else:
                self._parse_command(self._command.get())
                status = CONFIRMED
            if status == CONFIRMED:
                self._lbl_send_command.config(image=self.img_pending)
                self.set_status("Command(s) sent")
            elif status == CANCELLED:
                self.set_status("Command(s) cancelled")
            elif status == NOMINAL:
                self.set_status("Command(s) sent, no results")
            self._confirm = False

        except Exception as err:  # pylint: disable=broad-except
            self.set_status(f"Error {err}", ERRCOL)
            self._lbl_send_command.config(image=self.img_warn)

    def _parse_command(self, command: str):
        """
        Parse and send user-defined command(s).

        This could result in any number of errors if the
        ttypresets list contains garbage, so there's a broad
        catch-all-exceptions in the calling routine.

        :param str command: semicolon-delimited list of TTY commands
        """

        try:
            for cmd in command.split(";"):
                cmd = cmd.strip().encode(ASCII, errors=BSR)
                if self._crlf.get():
                    cmd += CRLF
                self.__app.gnss_outqueue.put(cmd)
                # self.logger.debug(f"command sent {cmd=}")
                if self._echo.get():  # echo output command to console
                    self.__app.gnss_inqueue.put((cmd, cmd.decode(ASCII, errors=BSR)))
                    self.__master.event_generate(TTY_EVENT)
        except Exception as err:  # pylint: disable=broad-except
            self.set_status(f"Error {err}", ERRCOL)
            self._lbl_send_command.config(image=self.img_warn)

    def update_status(self, msg: bytes):
        """
        Update pending confirmation status.

        :param bytes msg: ASCII config message
        """

        msgstr = msg.decode(ASCII, errors=BSR).upper()
        for ack in TTYOK:
            if ack in msgstr:
                self._lbl_send_command.config(image=self.img_confirmed)
                self.set_status("Command(s) acknowledged", OKCOL)
                return
        for nak in TTYERR:
            if nak in msgstr:
                self._lbl_send_command.config(image=self.img_warn)
                self.set_status("Command(s) rejected", ERRCOL)
                break



"""
widget_state.py

Class holding global constants, strings and dictionaries
used to maintain the state of the various user-selectable
widgets in the main container frame.

To add a new widget to PyGPSClient:
1. Create a new frame class `widgetname_frame.py`. The frame class
should implement `init_frame()` and `update_frame()` functions.
2. If the widget requires certain UBX messages to be enabled,
implement an `enable_messages(status)` function.
3. Add an entry to the foot of the self.__app.widget_state.state dictionary.
4. If the widget requires data not already in the `app.gnss_status`
data dictionary, add the requisite data items to the `GNSSStatus`
class definition and update `ubx_handler` to populate them.

Created on 30 Apr 2023

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import E, N, S, W

from pygpsclient.banner_frame import BannerFrame
from pygpsclient.chart_frame import ChartviewFrame
from pygpsclient.console_frame import ConsoleFrame
from pygpsclient.globals import CLASS, FRAME
from pygpsclient.graphview_frame import GraphviewFrame
from pygpsclient.imu_frame import IMUFrame
from pygpsclient.map_frame import MapviewFrame
from pygpsclient.rover_frame import RoverFrame
from pygpsclient.scatter_frame import ScatterViewFrame
from pygpsclient.settings_frame import SettingsFrame
from pygpsclient.skyview_frame import SkyviewFrame
from pygpsclient.spectrum_frame import SpectrumviewFrame
from pygpsclient.status_frame import StatusFrame
from pygpsclient.sysmon_frame import SysmonFrame

COL = "COL"
COLSPAN = "colspan"
DEFAULT = "def"
HIDE = "Hide"
MAXCOLSPAN = 4  # max no of widget columns
MAXROWSPAN = 4  # max no of widget rows
MENU = "men"
RESET = "rst"
ROW = "row"
ROWSPAN = "rowspan"
SHOW = "Show"
STICKY = "sty"
VISIBLE = "vis"
WDGBANNER = "Banner"
WDGCONSOLE = "Console"
WDGLEVELS = "Levels"
WDGMAP = "Map"
WDGROVER = "Rover Plot"
WDGSATS = "Satellites"
WDGSCATTER = "Scatter Plot"
WDGSETTINGS = "Settings"
WDGSPECTRUM = "Spectrum"
WDGSTATUS = "Status"
WDGSYSMON = "System Monitor"
WDGCHART = "Chart Plot"
WDGIMUMON = "IMU Monitor"


class WidgetState:
    """
    Class holding current state of all PyGPSClient widgets.
    """

    def __init__(self):
        """
        Constructor.
        """

        self.state = {
            # these widgets have fixed positions
            WDGBANNER: {  # always on top
                DEFAULT: True,
                MENU: False,
                CLASS: BannerFrame,
                FRAME: "frm_banner",
                VISIBLE: True,
                STICKY: (N, W, E, S),
                COL: 0,
                ROW: 0,
                COLSPAN: 6,
            },
            WDGSETTINGS: {  # always on right
                DEFAULT: True,
                CLASS: SettingsFrame,
                FRAME: "frm_settings",
                VISIBLE: True,
                STICKY: (N, W, E, S),
                COL: 5,
                ROW: 1,
                ROWSPAN: 4,
            },
            WDGSTATUS: {  # always on bottom
                DEFAULT: True,
                MENU: False,
                CLASS: StatusFrame,
                FRAME: "frm_status",
                VISIBLE: True,
                STICKY: (S, W, E),
                COL: 0,
                ROW: 5,
                COLSPAN: 6,
            },
            # these widgets rearrange dynamically according to
            # which has been selected to be visible
            WDGCONSOLE: {
                DEFAULT: True,
                CLASS: ConsoleFrame,
                FRAME: "frm_console",
                VISIBLE: True,
                COLSPAN: MAXCOLSPAN,
            },
            WDGSATS: {
                DEFAULT: True,
                CLASS: SkyviewFrame,
                FRAME: "frm_satview",
                VISIBLE: True,
            },
            WDGLEVELS: {
                DEFAULT: True,
                CLASS: GraphviewFrame,
                FRAME: "frm_graphview",
                VISIBLE: True,
            },
            WDGMAP: {
                DEFAULT: True,
                CLASS: MapviewFrame,
                FRAME: "frm_mapview",
                VISIBLE: True,
                RESET: True,
            },
            WDGSPECTRUM: {
                CLASS: SpectrumviewFrame,
                FRAME: "frm_spectrumview",
                VISIBLE: False,
                RESET: True,
            },
            WDGSCATTER: {
                CLASS: ScatterViewFrame,
                FRAME: "frm_scatterview",
                VISIBLE: False,
            },
            WDGROVER: {
                DEFAULT: False,
                CLASS: RoverFrame,
                FRAME: "frm_roverview",
                VISIBLE: False,
            },
            WDGCHART: {
                CLASS: ChartviewFrame,
                FRAME: "frm_chartview",
                VISIBLE: False,
                COLSPAN: 2,
            },
            WDGSYSMON: {
                CLASS: SysmonFrame,
                FRAME: "frm_sysmon",
                VISIBLE: False,
            },
            WDGIMUMON: {
                CLASS: IMUFrame,
                FRAME: "frm_imumon",
                VISIBLE: False,
            },
            # add any new widgets here
        }



"""
map_frame.py

Mapview frame class for PyGPSClient application.

This handles a frame containing a location map which can be either:

 - one or more fixed offline maps based on user-provided georeferenced
   images e.g. geoTIFF (defaults to Mercator world image).
 - dynamic online map or satellite image accessed via a MapQuest API.

NOTE: The free MapQuest API key is subject to a limit of 15,000
transactions / month, or roughly 500 / day, so the map updates are only
run periodically (once a minute). This utility is NOT intended to be used for
real time navigation.

Created on 13 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from time import time
from tkinter import (
    DISABLED,
    NORMAL,
    Checkbutton,
    E,
    Frame,
    IntVar,
    Label,
    N,
    S,
    Spinbox,
    StringVar,
    W,
)

from pygpsclient.globals import (
    BGCOL,
    ERRCOL,
    PNTCOL,
    READONLY,
    WIDGETU2,
    WORLD,
    Point,
)
from pygpsclient.map_canvas import HYB, MAP, MAPTYPES, SAT, TAG_CLOCK, MapCanvas
from pygpsclient.mapquest import (
    MAX_ZOOM,
    MIN_UPDATE_INTERVAL,
    MIN_ZOOM,
)
from pygpsclient.strings import LBLSHOWTRACK, NOWEBMAPFIX

ZOOMCOL = ERRCOL
ZOOMEND = "lightgray"
POSCOL = ERRCOL
TRK_COL = "magenta"  # color of track
INSET = 4


class MapviewFrame(Frame):
    """
    Map frame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)

        Frame.__init__(self, self.__master, *args, **kwargs)

        def_w, def_h = WIDGETU2
        self.width = kwargs.get("width", def_w)
        self.height = kwargs.get("height", def_h)
        self._img = None
        self._last_map_update = 0
        self._lastmaptype = ""
        self._lastmappath = ""
        self._bounds = None
        self._maptype = StringVar()
        self._maptype.set(self.__app.configuration.get("maptype_s"))
        self._showtrack = IntVar()
        self._showtrack.set(self.__app.configuration.get("showtrack_b"))
        self._mapzoom = IntVar()
        self._mapzoom.set(self.__app.configuration.get("mapzoom_n"))
        self._body()
        self._do_layout()
        self._reset()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self._can_mapview = MapCanvas(
            self.__app,
            self,
            height=self.height,
            width=self.width,
            bg=BGCOL,
        )
        self._frm_options = Frame(self, bg=BGCOL)
        self._spn_maptype = Spinbox(
            self._frm_options,
            values=MAPTYPES,
            width=7,
            wrap=True,
            textvariable=self._maptype,
            state=READONLY,
            background=BGCOL,
            readonlybackground=BGCOL,
            foreground=PNTCOL,
            buttonbackground=BGCOL,
        )
        self._lbl_zoom = Label(
            self._frm_options, width=5, text="Zoom", bg=BGCOL, fg=PNTCOL
        )
        self._spn_zoom = Spinbox(
            self._frm_options,
            from_=MIN_ZOOM,
            to=MAX_ZOOM,
            width=4,
            wrap=False,
            textvariable=self._mapzoom,
            state=READONLY,
            background=BGCOL,
            readonlybackground=BGCOL,
            disabledbackground=BGCOL,
            disabledforeground="grey",
            foreground=PNTCOL,
            buttonbackground=BGCOL,
        )
        self._chk_showtrack = Checkbutton(
            self._frm_options,
            text=LBLSHOWTRACK,
            variable=self._showtrack,
            background=BGCOL,
            foreground=PNTCOL,
        )

    def _do_layout(self):
        """
        Arrange widgets in frame.
        """

        self._can_mapview.grid(column=0, row=0, sticky=(N, S, E, W))
        self._frm_options.grid(column=0, row=1, sticky=(E, W))
        self._spn_maptype.grid(column=0, row=0, padx=1, sticky=W)
        self._lbl_zoom.grid(column=1, row=0, padx=1, sticky=W)
        self._spn_zoom.grid(column=2, row=0, padx=1, sticky=W)
        self._chk_showtrack.grid(column=3, row=0, padx=1, sticky=W)
        self.grid_columnconfigure(0, weight=1)
        self.grid_rowconfigure(0, weight=1)
        self._can_mapview.grid_columnconfigure(0, weight=1)
        self._can_mapview.grid_rowconfigure(0, weight=1)

    def _attach_events(self):
        """
        Bind events to frame.
        """

        self.bind("<Configure>", self._on_resize)
        self._maptype.trace_add("write", self.on_maptype)
        self._mapzoom.trace_add("write", self.on_mapzoom)
        self._showtrack.trace_add("write", self.on_showtrack)
        self._can_mapview.bind("<Double-Button-1>", self.on_refresh)
        self._lbl_zoom.bind(
            "<Double-Button-1>",
            lambda event, val=0: self._setzoom(val),
        )
        self._lbl_zoom.bind(
            "<Double-Button-2>",
            lambda event, val=MAX_ZOOM: self._setzoom(val),
        )
        self._lbl_zoom.bind(
            "<Double-Button-3>",
            lambda event, val=MAX_ZOOM: self._setzoom(val),
        )

    def _reset(self):
        """
        Initialise mapview panel.
        """

        self._maptype.set(self.__app.configuration.get("maptype_s"))
        self.on_maptype(None, None, None)
        self._mapzoom.set(self.__app.configuration.get("mapzoom_n"))
        self._showtrack.set(self.__app.configuration.get("showtrack_b"))

    def on_refresh(self, event):  # pylint: disable=unused-argument
        """
        Trigger refresh of web map.

        :param event: event
        """

        self._last_map_update = 0

    def on_maptype(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Set maptype event binding.
        """

        if self._maptype.get() == self._lastmaptype:
            return

        if self._maptype.get() == WORLD:
            self._lbl_zoom.config(state=DISABLED)
            self._spn_zoom.config(state=DISABLED)
        else:
            self._lbl_zoom.config(state=NORMAL)
            self._spn_zoom.config(state=READONLY)

        self.__app.configuration.set("maptype_s", self._maptype.get())
        self._lastmaptype = self._maptype.get()
        self.on_refresh(None)

    def on_mapzoom(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Set mapzoom event binding.
        """

        self.__app.configuration.set("mapzoom_n", self._mapzoom.get())
        self.on_refresh(None)

    def _setzoom(self, val: int = 0):
        """
        Set zoom level.

        :param int val: zoom value
        """

        if val < 1:
            val = int((MAX_ZOOM + MIN_ZOOM) / 2)
        self._mapzoom.set(val)

    def on_showtrack(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Set show track event binding.
        """

        self.__app.configuration.set("showtrack_b", self._showtrack.get())
        if not self._showtrack.get():
            self._can_mapview.track = None
        self.on_refresh(None)

    def update_frame(self):
        """
        Draw map and mark current known position and horizontal accuracy (where available).
        """

        lat = self.__app.gnss_status.lat
        lon = self.__app.gnss_status.lon
        hacc = self.__app.gnss_status.hacc
        map_update_interval = max(
            self.__app.configuration.get("mapupdateinterval_n"),
            MIN_UPDATE_INTERVAL,
        )
        self._maptype.set(self.__app.configuration.get("maptype_s"))

        # if no valid position, display warning message
        # fix = kwargs.get("fix", 0)
        if (
            lat in (None, "")
            or lon in (None, "")
            or (lat == 0 and lon == 0)
            # or fix in (0, 5)  # no fix or time only
        ):
            self.reset_map_refresh()
            self._can_mapview.draw_msg(NOWEBMAPFIX)
            return

        # record track if Show Track checkbox ticked
        if self._showtrack.get():
            self._can_mapview.track = Point(lat, lon)

        # limit mapquest calls to specified interval to avoid cost
        maptype = self._maptype.get()
        if maptype in (MAP, SAT, HYB):
            now = time()
            if now - self._last_map_update < map_update_interval:
                self._can_mapview.draw_countdown(
                    (-360 / map_update_interval) * (now - self._last_map_update)
                )
                return
            self._last_map_update = now
        else:
            self._can_mapview.delete(TAG_CLOCK)

        self._can_mapview.draw_map(
            maptype=self._maptype.get(),
            location=Point(lat, lon),
            track=self._can_mapview.track,
            marker=self._can_mapview.marker,
            hacc=hacc,
            zoom=self._mapzoom.get(),
        )
        self._bounds = self._can_mapview.bounds

        if self._can_mapview.zoommin:
            self._spn_zoom.config(highlightbackground=ERRCOL, highlightthickness=3)
        else:
            self._spn_zoom.config(highlightbackground="gray90", highlightthickness=3)

    def reset_map_refresh(self):
        """
        Reset map refresh counter to zero
        """

        self._last_map_update = 0

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame

        :param event event: resize event
        """

        self.width, self.height = self.get_size()

    def get_size(self):
        """
        Get current canvas size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about any resizing
        return self._can_mapview.winfo_width(), self._can_mapview.winfo_height()



"""
app.py

PyGPSClient - Main tkinter application class.

- Loads configuration from json file (if available)
- Instantiates all frames, widgets, and protocol handlers.
- Starts and stops threaded dialog and protocol handler processes.
- Maintains current serial and RTK connection status.
- Reacts to various message events, processes navigation data
  placed on input message queue by serial, socket or file stream reader
  and assigns to appropriate NMEA, UBX or RTCM protocol handler.
- Maintains central dictionary of current key navigation data as
  `gnss_status`, for use by user-selectable widgets.

Global logging configuration is defined in __main__.py. To enable module
logging, this and other subsidiary modules can use:

```self.logger = logging.getLogger(__name__)```

To override individual module loglevel, use e.g.

```self.logger.setLevel(INFO)```

Created on 12 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

# pylint: disable=too-many-ancestors, no-member

import logging
from datetime import datetime, timedelta
from queue import Empty, Queue
from threading import Thread
from tkinter import E, Frame, N, PhotoImage, S, Tk, Toplevel, W, font

from pygnssutils import GNSSMQTTClient, GNSSNTRIPClient, MQTTMessage
from pygnssutils.socket_server import ClientHandler, SocketServer
from pynmeagps import NMEAMessage
from pyrtcm import RTCMMessage
from pysbf2 import SBFMessage
from pyspartn import SPARTNMessage
from pyubx2 import NMEA_PROTOCOL, POLL, RTCM3_PROTOCOL, UBX_PROTOCOL, UBXMessage
from serial import SerialException, SerialTimeoutException

from pygpsclient._version import __version__ as VERSION
from pygpsclient.configuration import Configuration
from pygpsclient.dialog_state import DialogState
from pygpsclient.file_handler import FileHandler
from pygpsclient.globals import (
    CFG,
    CLASS,
    CONFIGFILE,
    CONNECTED,
    CONNECTED_NTRIP,
    CONNECTED_SPARTNIP,
    CONNECTED_SPARTNLB,
    DISCONNECTED,
    ERRCOL,
    FRAME,
    GNSS_EOF_EVENT,
    GNSS_ERR_EVENT,
    GNSS_EVENT,
    GNSS_TIMEOUT_EVENT,
    ICON_APP128,
    INFOCOL,
    MQTT_PROTOCOL,
    NOPORTS,
    NTRIP_EVENT,
    OKCOL,
    SBF_PROTOCOL,
    SOCKSERVER_MAX_CLIENTS,
    SPARTN_EVENT,
    SPARTN_PROTOCOL,
    THD,
    TTY_EVENT,
    TTY_PROTOCOL,
    TTYMARKER,
)
from pygpsclient.gnss_status import GNSSStatus
from pygpsclient.helpers import check_latest
from pygpsclient.menu_bar import MenuBar
from pygpsclient.nmea_handler import NMEAHandler
from pygpsclient.rtcm3_handler import RTCM3Handler
from pygpsclient.sbf_handler import SBFHandler
from pygpsclient.sqllite_handler import DBINMEM, SQLOK, SqliteHandler
from pygpsclient.stream_handler import StreamHandler
from pygpsclient.strings import (
    CONFIGERR,
    DLG,
    DLGSTOPRTK,
    DLGTNTRIP,
    ENDOFFILE,
    INACTIVE_TIMEOUT,
    INTROTXTNOPORTS,
    LOADCONFIGBAD,
    LOADCONFIGNONE,
    LOADCONFIGOK,
    NOTCONN,
    NOWDGSWARN,
    SAVECONFIGBAD,
    SAVECONFIGOK,
    TITLE,
    VERCHECK,
)
from pygpsclient.tty_handler import TTYHandler
from pygpsclient.ubx_handler import UBXHandler
from pygpsclient.widget_state import (
    COL,
    COLSPAN,
    DEFAULT,
    HIDE,
    MAXCOLSPAN,
    MAXROWSPAN,
    MENU,
    ROW,
    ROWSPAN,
    SHOW,
    STICKY,
    VISIBLE,
    WDGCHART,
    WDGCONSOLE,
    WidgetState,
)


class App(Frame):
    """
    Main PyGPSClient GUI Application Class.
    """

    def __init__(self, master, **kwargs):  # pylint: disable=too-many-statements
        """
        Set up main application and add frames.

        :param tkinter.Tk master: reference to Tk root
        :param args: optional args
        :param kwargs: optional (CLI) kwargs
        """

        self.__master = master
        self.logger = logging.getLogger(__name__)

        # Init Frame class
        Frame.__init__(self, self.__master)

        self.__master.protocol("WM_DELETE_WINDOW", self.on_exit)
        self.__master.title(TITLE)
        self.__master.iconphoto(True, PhotoImage(file=ICON_APP128))

        self.gnss_inqueue = Queue()  # messages from GNSS receiver
        self.gnss_outqueue = Queue()  # messages to GNSS receiver
        self.ntrip_inqueue = Queue()  # messages from NTRIP source
        self.spartn_inqueue = Queue()  # messages from SPARTN correction rcvr
        self.spartn_outqueue = Queue()  # messages to SPARTN correction rcvr
        self.socket_inqueue = Queue()  # message from socket
        self.socket_outqueue = Queue()  # message to socket
        self.widget_state = WidgetState()  # widget state
        self.dialog_state = DialogState()  # dialog state
        self.configuration = Configuration(self)  # configuration state
        self.gnss_status = GNSSStatus()  # holds latest GNSS readings
        self.file_handler = FileHandler(self)
        self.stream_handler = StreamHandler(self)
        self.spartn_stream_handler = StreamHandler(self)
        self.nmea_handler = NMEAHandler(self)
        self.ubx_handler = UBXHandler(self)
        self.sbf_handler = SBFHandler(self)
        self.rtcm_handler = RTCM3Handler(self)
        self.tty_handler = TTYHandler(self)
        self.ntrip_handler = GNSSNTRIPClient(self)
        self.spartn_handler = GNSSMQTTClient(self)
        self.sqlite_handler = SqliteHandler(self)
        self._conn_status = DISCONNECTED
        self._rtk_conn_status = DISCONNECTED
        self._nowidgets = True
        self._last_gui_update = datetime.now()
        self._socket_thread = None
        self._socket_server = None
        self._colcount = 0
        self._rowcount = 0
        self._consoledata = []
        # check if platform supports spatialite database
        self.db_enabled = self.sqlite_handler.open(dbname=DBINMEM)

        # load config from json file
        configfile = kwargs.pop("config", CONFIGFILE)
        _, configerr = self.configuration.loadfile(configfile)
        # load config from CLI arguments & env variables
        self.configuration.loadcli(**kwargs)
        if configerr == "":
            self.update_widgets()  # set initial widget state

        self._body()
        self._do_layout()
        self._attach_events()

        # display config load status once status frame has been instantiated
        if configerr == "":
            if self._nowidgets:  # if all widgets have been disabled in config
                self.set_status(NOWDGSWARN.format(configfile), ERRCOL)
            else:
                self.set_status(LOADCONFIGOK.format(configfile), OKCOL)
        else:
            if "No such file or directory" in configerr:
                self.set_status(LOADCONFIGNONE.format(configfile), ERRCOL)
            else:
                self.set_status(LOADCONFIGBAD.format(configfile, configerr), ERRCOL)

        # initialise widgets
        for value in self.widget_state.state.values():
            frm = getattr(self, value[FRAME])
            if hasattr(frm, "init_frame"):
                frm.init_frame()

        # display initial connection status
        self.frm_banner.update_conn_status(DISCONNECTED)
        if self.frm_settings.frm_serial.status == NOPORTS:
            self.set_status(INTROTXTNOPORTS, ERRCOL)

        # open database if database recording enabled
        dbpath = self.configuration.get("databasepath_s")
        if (
            self.db_enabled == SQLOK
            and self.configuration.get("database_b")
            and dbpath != ""
        ):
            rc = self.sqlite_handler.open(dbpath=dbpath)
            self.configuration.set("database_b", rc == SQLOK)

        # check for more recent version (if enabled)
        if self.configuration.get("checkforupdate_b") and configerr == "":
            self._check_update()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self._set_default_fonts()
        self.menu = MenuBar(self)
        self.__master.config(menu=self.menu)

        # instantiate widgets
        for value in self.widget_state.state.values():
            setattr(
                self,
                value[FRAME],
                value[CLASS](self, borderwidth=2, relief="groove"),
            )

    def _do_layout(self):
        """
        Arrange widgets in main application frame, and set
        widget visibility and menu label (show/hide).
        """

        col = 0
        row = 1
        maxcol = 0
        maxrow = 0
        men = 0
        for name in self.widget_state.state:
            col, row, maxcol, maxrow, men = self._widget_grid(
                name, col, row, maxcol, maxrow, men
            )

        # ensure widgets expand to size of container (needed
        # when not using 'pack' grid management)
        # weight = 0 means fixed, non-expandable
        # weight > 0 means expandable
        for col in range(MAXCOLSPAN + 1):
            self.__master.grid_columnconfigure(col, weight=0)
        for row in range(MAXROWSPAN + 2):
            self.__master.grid_rowconfigure(row, weight=0)
        # print(f"{maxcol=} {maxrow=}")
        for col in range(maxcol):
            self.__master.grid_columnconfigure(col, weight=5)
        for row in range(1, maxrow + 1):
            self.__master.grid_rowconfigure(row, weight=5)

    def _widget_grid(
        self, name: str, col: int, row: int, maxcol: int, maxrow: int, men: int
    ) -> tuple:
        """
        Arrange widgets and update menu label (show/hide).

        Widgets with explicit COL settings will be placed in fixed
        positions; widgets with no COL setting will be arranged
        dynamically.

        :param str name: name of widget
        :param int col: col
        :param int row: row
        :param int maxcol: max cols
        :param int maxrow: max rows
        :param int men: menu position
        :return: max row & col
        :rtype: tuple
        """

        wdg = self.widget_state.state[name]
        dynamic = wdg.get(COL, None) is None
        frm = getattr(self, wdg[FRAME])
        if wdg[VISIBLE]:
            self.widget_enable_messages(name)
            fcol = wdg.get(COL, col)
            frow = wdg.get(ROW, row)
            colspan = wdg.get(COLSPAN, 1)
            rowspan = wdg.get(ROWSPAN, 1)
            if dynamic and fcol + colspan > MAXCOLSPAN:
                fcol = 0
                frow += 1
            frm.grid(
                column=fcol,
                row=frow,
                columnspan=colspan,
                rowspan=rowspan,
                padx=2,
                pady=2,
                sticky=wdg.get(STICKY, (N, S, W, E)),
            )
            lbl = HIDE
            if dynamic:
                col += colspan
                if col >= MAXCOLSPAN:
                    col = 0
                    row += rowspan
                maxcol = max(maxcol, fcol + colspan)
                maxrow = max(maxrow, frow)
        else:
            frm.grid_forget()
            lbl = SHOW

        # update menu label (show/hide)
        if wdg.get(MENU, True):
            self.menu.view_menu.entryconfig(men, label=f"{lbl} {name}")
            men += 1

        # force widget to rescale
        frm.event_generate("<Configure>")

        return col, row, maxcol, maxrow, men

    def widget_toggle(self, name: str):
        """
        Toggle widget visibility and enable or disable any
        UBX messages required by widget.

        :param str name: widget name
        """

        self.widget_state.state[name][VISIBLE] = not self.widget_state.state[name][
            VISIBLE
        ]
        self.configuration.set(name, self.widget_state.state[name][VISIBLE])
        self._do_layout()

    def widget_enable_messages(self, name: str):
        """
        Enable any NMEA, UBX or RTCM messages required by widget.

        :param str name: widget name
        """

        frm = getattr(self, self.widget_state.state[name][FRAME])
        if hasattr(frm, "enable_messages"):
            frm.enable_messages(self.widget_state.state[name][VISIBLE])

    def widget_reset(self):
        """
        Reset widgets to default layout.
        """

        for nam, wdg in self.widget_state.state.items():
            vis = wdg.get(DEFAULT, False)
            wdg[VISIBLE] = vis
            self.configuration.set(nam, vis)
        self._do_layout()

    def reset_gnssstatus(self):
        """
        Reset gnss_status dict e.g. after reconnecting.
        """

        self.gnss_status = GNSSStatus()

    def _attach_events(self):
        """
        Bind events to main application.
        """

        self.__master.bind(GNSS_EVENT, self.on_gnss_read)
        self.__master.bind(GNSS_EOF_EVENT, self.on_gnss_eof)
        self.__master.bind(GNSS_TIMEOUT_EVENT, self.on_gnss_timeout)
        self.__master.bind(GNSS_ERR_EVENT, self.on_stream_error)
        self.__master.bind(NTRIP_EVENT, self.on_ntrip_read)
        self.__master.bind(SPARTN_EVENT, self.on_spartn_read)
        self.__master.bind(TTY_EVENT, self.on_tty_read)
        self.__master.bind_all("<Control-q>", self.on_exit)

    def _set_default_fonts(self):
        """
        Set default fonts for entire application.
        """
        # pylint: disable=attribute-defined-outside-init

        self.font_vsm = font.Font(size=8)
        self.font_sm = font.Font(size=10)
        self.font_md = font.Font(size=12)
        self.font_md2 = font.Font(size=14)
        self.font_lg = font.Font(size=18)

    def set_connection(self, message, color=OKCOL):
        """
        Sets connection description in status bar.

        :param str message: message to be displayed in connection label
        :param str color: rgb color string

        """

        if hasattr(self, "frm_status"):
            self.frm_status.set_connection(message, color=color)

    def set_status(self, message, color=OKCOL):
        """
        Sets text of status bar.

        :param str message: message to be displayed in status label
        :param str color: rgb color string

        """

        if hasattr(self, "frm_status"):
            color = INFOCOL if color == "blue" else color
            self.frm_status.set_status(message, color)
            self.update_idletasks()

    def set_event(self, evt: str):
        """
        Generate event

        :param str evt: event type string
        """

        self.__master.event_generate(evt)

    def load_config(self):
        """
        Load configuration file menu option.
        """

        # Warn if Streaming, NTRIP or SPARTN clients are running
        if self.conn_status == DISCONNECTED and self.rtk_conn_status == DISCONNECTED:
            self.set_status("", OKCOL)
        else:
            self.set_status(DLGSTOPRTK, ERRCOL)
            return

        filename, err = self.configuration.loadfile()
        if err == "":  # load succeeded
            self.update_widgets()
            for frm in (
                self.frm_settings,
                self.frm_settings.frm_serial,
                self.frm_settings.frm_socketclient,
                self.frm_settings.frm_socketserver,
            ):
                frm.reset()
            self._do_layout()
            if self._nowidgets:
                self.set_status(NOWDGSWARN.format(filename), ERRCOL)
            else:
                self.set_status(LOADCONFIGOK.format(filename), OKCOL)
        elif err == "cancelled":  # user cancelled
            return
        else:  # config error
            self.set_status(LOADCONFIGBAD.format(filename), ERRCOL)

    def save_config(self):
        """
        Save configuration file menu option.
        """

        err = self.configuration.savefile()
        if err == "":
            self.set_status(SAVECONFIGOK, OKCOL)
        else:  # save failed
            self.set_status(SAVECONFIGBAD.format(err), ERRCOL)

    def update_widgets(self):
        """
        Update widget configuration (self.widget_state.state).

        If no widgets are set as "visible" in the config file, enable
        the status widget anyway to display a warning to this effect.
        """

        try:
            self._nowidgets = True
            for key, vals in self.widget_state.state.items():
                vis = self.configuration.get(key)
                vals[VISIBLE] = vis
                if vis:
                    self._nowidgets = False
            if self._nowidgets:
                self.widget_state.state["Status"][VISIBLE] = "true"
        except KeyError as err:
            self.set_status(f"{CONFIGERR} - {err}", ERRCOL)

    def start_dialog(self, dlg: str):
        """
        Start a threaded dialog task if the dialog is not already open.

        :param str dlg: name of dialog
        """

        if self.dialog_state.state[dlg][THD] is None:
            self.dialog_state.state[dlg][THD] = Thread(
                target=self._dialog_thread, args=(dlg,), daemon=False
            )
            self.dialog_state.state[dlg][THD].start()

    def _dialog_thread(self, dlg: str):
        """
        THREADED PROCESS

        Dialog thread.

        :param str dlg: name of dialog
        """

        config = (
            self.configuration.settings if self.dialog_state.state[dlg][CFG] else {}
        )
        cls = self.dialog_state.state[dlg][CLASS]
        self.dialog_state.state[dlg][DLG] = cls(self, saved_config=config)

    def stop_dialog(self, dlg: str):
        """
        Register dialog as closed.

        :param str dlg: name of dialog
        """

        self.dialog_state.state[dlg][THD] = None
        self.dialog_state.state[dlg][DLG] = None

    def dialog(self, dlg: str) -> Toplevel:
        """
        Get reference to dialog instance.

        :param str dlg: name of dialog
        :return: dialog instance
        :rtype: Toplevel
        """

        return self.dialog_state.state[dlg][DLG]

    def start_sockserver_thread(self):
        """
        Start socket server thread.
        """

        cfg = self.configuration
        host = cfg.get("sockhost_s")
        port = cfg.get("sockport_n")
        ntripmode = cfg.get("sockmode_b")
        ntripuser = cfg.get("ntripcasteruser_s")
        ntrippassword = cfg.get("ntripcasterpassword_s")
        self._socket_thread = Thread(
            target=self._sockserver_thread,
            args=(
                ntripmode,
                host,
                port,
                ntripuser,
                ntrippassword,
                SOCKSERVER_MAX_CLIENTS,
                self.socket_outqueue,
            ),
            daemon=True,
        )
        self._socket_thread.start()
        self.frm_banner.update_transmit_status(0)

    def stop_sockserver_thread(self):
        """
        Stop socket server thread.
        """

        self.frm_banner.update_transmit_status(-1)
        if self._socket_server is not None:
            self._socket_server.shutdown()

    def _sockserver_thread(
        self,
        ntripmode: int,
        host: str,
        port: int,
        ntripuser: str,
        ntrippassword: str,
        maxclients: int,
        socketqueue: Queue,
    ):
        """
        THREADED
        Socket Server thread.

        :param int ntripmode: 0 = open socket server, 1 = NTRIP server
        :param str host: socket host name (0.0.0.0)
        :param int port: socket port (50010)
        :param int maxclients: max num of clients (5)
        :param Queue socketqueue: socket server read queue
        """

        try:
            with SocketServer(
                self,
                ntripmode,
                maxclients,
                socketqueue,
                (host, port),
                ClientHandler,
                ntripuser=ntripuser,
                ntrippassword=ntrippassword,
            ) as self._socket_server:
                self._socket_server.serve_forever()
        except OSError as err:
            self.set_status(f"Error starting socket server {err}", ERRCOL)

    def update_clients(self, clients: int):
        """
        Update number of connected clients in settings panel.

        :param int clients: no of connected clients
        """

        self.frm_settings.frm_socketserver.clients = clients

    def on_exit(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Kill any running processes and quit application.
        """

        self.file_handler.close_logfile()
        self.file_handler.close_trackfile()
        self.stop_sockserver_thread()
        self.stream_handler.stop_read_thread()
        self.sqlite_handler.close()
        self.__master.destroy()

    def on_gnss_read(self, event):  # pylint: disable=unused-argument
        """
        EVENT TRIGGERED
        Action on <<gnss_read>> event - data available on GNSS queue.

        :param event event: read event
        """

        try:
            raw_data, parsed_data = self.gnss_inqueue.get(False)
            if raw_data is not None and parsed_data is not None:
                self.process_data(raw_data, parsed_data)
            # if socket server is running, output raw data to socket
            if self.frm_settings.frm_socketserver.socketserving:
                self.socket_outqueue.put(raw_data)
            self.gnss_inqueue.task_done()
        except Empty:
            pass

    def on_gnss_eof(self, event):  # pylint: disable=unused-argument
        """
        EVENT TRIGGERED
        Action on <<sgnss_eof>> event - end of file.

        :param event event: <<gnss_eof>> event
        """

        self.frm_settings.frm_socketserver.socketserving = (
            False  # turn off socket server
        )
        self._refresh_widgets()
        self.conn_status = DISCONNECTED
        self.set_status(ENDOFFILE, ERRCOL)

    def on_gnss_timeout(self, event):  # pylint: disable=unused-argument
        """
        EVENT TRIGGERED
        Action on <<sgnss_timeout>> event - stream inactivity timeout.

        :param event event: <<gnss_timeout>> event
        """

        self.frm_settings.frm_socketserver.socketserving = (
            False  # turn off socket server
        )
        self._refresh_widgets()
        self.conn_status = DISCONNECTED
        self.set_status(INACTIVE_TIMEOUT, ERRCOL)

    def on_stream_error(self, event):  # pylint: disable=unused-argument
        """
        EVENT TRIGGERED
        Action on "<<gnss_error>>" event - connection streaming error.

        :param event event: <<gnss_error>> event
        """

        self._refresh_widgets()
        self.conn_status = DISCONNECTED

    def on_ntrip_read(self, event):  # pylint: disable=unused-argument
        """
        EVENT TRIGGERED
        Action on <<ntrip_read>> event - data available on NTRIP queue.

        :param event event: read event
        """

        try:
            raw_data, parsed_data = self.ntrip_inqueue.get(False)
            if (
                raw_data is not None
                and parsed_data is not None
                and isinstance(raw_data, bytes)
            ):
                if self._rtk_conn_status == CONNECTED_NTRIP:
                    source = "NTRIP"
                else:
                    source = "OTHER"
                if isinstance(parsed_data, (RTCMMessage, SPARTNMessage)):
                    if self.conn_status == CONNECTED:
                        self.gnss_outqueue.put(raw_data)
                    self.process_data(raw_data, parsed_data, source + ">>")
                elif isinstance(parsed_data, NMEAMessage):
                    # i.e. NMEA GGA sentence sent to NTRIP server
                    self.process_data(raw_data, parsed_data, source + "<<")
            self.ntrip_inqueue.task_done()
        except Empty:
            pass
        except (SerialException, SerialTimeoutException) as err:
            self.set_status(f"Error sending to device {err}", ERRCOL)

    def on_spartn_read(self, event):  # pylint: disable=unused-argument
        """
        EVENT TRIGGERED
        Action on <<spartn_read>> event - data available on SPARTN queue.

        :param event event: read event
        """

        try:
            raw_data, parsed_data = self.spartn_inqueue.get(False)
            if raw_data is not None and parsed_data is not None:
                if self.conn_status == CONNECTED:
                    self.gnss_outqueue.put(raw_data)
                if self._rtk_conn_status == CONNECTED_SPARTNLB:
                    source = "LBAND"
                elif self._rtk_conn_status == CONNECTED_SPARTNIP:
                    source = "MQTT"
                else:
                    source = "OTHER"
                self.process_data(
                    raw_data,
                    parsed_data,
                    source + ">>",
                )
            self.spartn_inqueue.task_done()

        except Empty:
            pass
        except (SerialException, SerialTimeoutException) as err:
            self.set_status(f"Error sending to device {err}", ERRCOL)

    def on_tty_read(self, event):  # pylint: disable=unused-argument
        """
        EVENT TRIGGERED
        Action on <<tty_read>> event - data available on TTY queue.

        :param event event: read event
        """

        try:
            raw_data, parsed_data = self.gnss_inqueue.get(False)
            if raw_data is not None:
                self.process_data(raw_data, parsed_data, TTYMARKER)
            self.gnss_inqueue.task_done()
        except Empty:
            pass

    def update_ntrip_status(self, status: bool, msgt: tuple = None):
        """
        Update NTRIP configuration dialog connection status.

        :param bool status: connected to NTRIP server yes/no
        :param tuple msgt: tuple of (message, color)
        """

        if self.dialog(DLGTNTRIP) is not None:
            self.dialog(DLGTNTRIP).set_controls(status, msgt)

    def get_coordinates(self) -> dict:
        """
        Get current coordinates and fix data.

        :return: dict of coords and fix data
        :rtype: dict
        """

        try:
            sep = self.gnss_status.hae - self.gnss_status.alt
        except TypeError:
            sep = 0
        return {
            "connection": self._conn_status,
            "lat": self.gnss_status.lat,
            "lon": self.gnss_status.lon,
            "alt": self.gnss_status.alt,  # hmsl
            "hae": self.gnss_status.hae,
            "sep": sep,
            "sip": self.gnss_status.sip,
            "fix": self.gnss_status.fix,
            "hdop": self.gnss_status.hdop,
            "diffage": self.gnss_status.diff_age,
            "diffstation": self.gnss_status.diff_station,
        }

    def process_data(self, raw_data: bytes, parsed_data: object, marker: str = ""):
        """
        Update the various GUI widgets, GPX track and log file.

        :param bytes raw_data: raw message data
        :param object parsed data: NMEAMessage, UBXMessage or RTCMMessage
        :param str marker: string prepended to console entries e.g. "NTRIP>>"
        """

        # self.logger.debug(f"data received {parsed_data.identity}")
        msgprot = 0
        protfilter = self.configuration.get("protocol_n")
        if isinstance(parsed_data, NMEAMessage):
            msgprot = NMEA_PROTOCOL
        elif isinstance(parsed_data, SBFMessage):
            msgprot = SBF_PROTOCOL
        elif isinstance(parsed_data, UBXMessage):
            msgprot = UBX_PROTOCOL
        elif isinstance(parsed_data, RTCMMessage):
            msgprot = RTCM3_PROTOCOL
        elif isinstance(parsed_data, SPARTNMessage):
            msgprot = SPARTN_PROTOCOL
        elif isinstance(parsed_data, MQTTMessage):
            msgprot = MQTT_PROTOCOL
        elif isinstance(parsed_data, str):
            if self.configuration.get("ttyprot_b"):
                msgprot = TTY_PROTOCOL
            else:
                marker = "WARNING>>"

        if msgprot == UBX_PROTOCOL and msgprot & protfilter:
            self.ubx_handler.process_data(raw_data, parsed_data)
        elif msgprot == SBF_PROTOCOL and msgprot & protfilter:
            self.sbf_handler.process_data(raw_data, parsed_data)
        elif msgprot == NMEA_PROTOCOL and msgprot & protfilter:
            self.nmea_handler.process_data(raw_data, parsed_data)
        elif msgprot == RTCM3_PROTOCOL and msgprot & protfilter:
            self.rtcm_handler.process_data(raw_data, parsed_data)
        elif msgprot == TTY_PROTOCOL and msgprot & protfilter:
            self.tty_handler.process_data(raw_data, parsed_data)
        elif msgprot == SPARTN_PROTOCOL and msgprot & protfilter:
            pass
        elif msgprot == MQTT_PROTOCOL:
            pass

        # update chart data if chart is visible
        if self.widget_state.state[WDGCHART][VISIBLE]:
            getattr(self, self.widget_state.state[WDGCHART][FRAME]).update_data(
                parsed_data
            )

        # update consoledata if console is visible and protocol not filtered
        if self.widget_state.state[WDGCONSOLE][VISIBLE] and (
            msgprot in (0, MQTT_PROTOCOL) or msgprot & protfilter
        ):
            self._consoledata.append((raw_data, parsed_data, marker))

        # periodically update widgets if visible
        if datetime.now() > self._last_gui_update + timedelta(
            seconds=self.configuration.get("guiupdateinterval_f")
        ):
            self._refresh_widgets()
            # update database if enabled
            if self.configuration.get("database_b"):
                self.sqlite_handler.load_data()
            self._last_gui_update = datetime.now()

        # update GPX track file if enabled
        if self.configuration.get("recordtrack_b"):
            self.file_handler.update_gpx_track()

        # update log file if enabled
        if self.configuration.get("datalog_b"):
            self.file_handler.write_logfile(raw_data, parsed_data)

    def _refresh_widgets(self):
        """
        Refresh visible widgets.
        """

        for wdg, wdgdata in self.widget_state.state.items():
            frm = getattr(self, wdgdata[FRAME])
            if hasattr(frm, "update_frame") and wdgdata[VISIBLE]:
                if wdg == WDGCONSOLE:
                    frm.update_frame(self._consoledata)
                    self._consoledata = []
                else:
                    frm.update_frame()

    def _check_update(self):
        """
        Check for updated version.
        """

        latest = check_latest(TITLE)
        if latest not in (VERSION, "N/A"):
            self.set_status(f"{VERCHECK} {latest}", ERRCOL)

    def poll_version(self, protocol: int):
        """
        Poll hardware information message for device hardware & firmware version.

        :param int protocol: protocol(s)
        """

        msg = None
        if protocol & UBX_PROTOCOL:
            msg = UBXMessage("MON", "MON-VER", POLL)
        elif protocol & SBF_PROTOCOL:
            msg = b"SSSSSSSSSS\r\nesoc, COM1, ReceiverSetup\r\n"
        elif protocol & NMEA_PROTOCOL:
            msg = NMEAMessage("P", "QTMVERNO", POLL)

        if isinstance(msg, (UBXMessage, NMEAMessage)):
            self.gnss_outqueue.put(msg.serialize())
            self.set_status(
                f"{msg.identity} POLL message sent",
            )
        elif isinstance(msg, bytes):
            self.gnss_outqueue.put(msg)
            self.set_status(
                "Setup POLL message sent",
            )

    @property
    def appmaster(self) -> Tk:
        """
        Getter for application master (Tk).

        :return: reference to master Tk instance
        :rtype: Tk
        """

        return self.__master

    @property
    def conn_status(self) -> int:
        """
        Getter for connection status.

        :return: connection status e.g. 1 = CONNECTED
        :rtype: int
        """

        return self._conn_status

    @conn_status.setter
    def conn_status(self, status: int):
        """
        Setter for connection status.

        :param int status: connection status e.g. 1 = CONNECTED
        """

        self._conn_status = status
        self.frm_banner.update_conn_status(status)
        self.frm_settings.enable_controls(status)
        if status == DISCONNECTED:
            self.set_connection(NOTCONN)

    @property
    def rtk_conn_status(self) -> int:
        """
        Getter for SPARTN connection status.

        :return: connection status
        :rtype: int
        """

        return self._rtk_conn_status

    @rtk_conn_status.setter
    def rtk_conn_status(self, status: int):
        """
        Setter for SPARTN connection status.

        :param int status: connection status
        """

        self._rtk_conn_status = status
        self.frm_banner.update_rtk_status(status)



"""
status_frane.py

Status Bar frame class for PyGPSClient application.

This handles the status bar notifications at the foot of the window.

Created on 12 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import VERTICAL, E, Frame, Label, N, S, StringVar, W, ttk


class StatusFrame(Frame):
    """
    Status bar frame class.
    """

    def __init__(self, app, *args, **kwargs):
        """
        Constructor

        :param Frame app: reference to main tkinter application
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        Frame.__init__(self, self.__master, *args, **kwargs)

        self._status = StringVar()
        self._connection = StringVar()
        self.width, self.height = self.get_size()
        self._body()

        self.bind("<Configure>", self._on_resize)

    def _body(self):
        """
        Set up frame and widgets.
        """

        self.grid_rowconfigure(0, weight=1)

        self.option_add("*Font", self.__app.font_md)

        self._lbl_connection = Label(self, textvariable=self._connection, anchor=W)
        self._lbl_status_preset = Label(self, textvariable=self._status, anchor=W)
        self._lbl_connection.grid(column=0, row=0, sticky=(W, E))
        ttk.Separator(self, orient=VERTICAL).grid(column=1, row=0, sticky=(N, S))
        self._lbl_status_preset.grid(column=2, row=0, sticky=(W, E))

    def set_connection(self, connection: str, color: str = ""):
        """
        Sets connection description in status bar.

        :param str connection: description of connection
        :param str color: rgb color string (default=blue)

        """

        if len(connection) > 100:
            connection = "..." + connection[-100:]
        if color != "":
            self._lbl_connection.config(fg=color)
        self._connection.set("  " + connection)

    def set_status(self, message, color: str = ""):
        """
        Sets message in status bar.

        :param str message: message to be displayed in status bar
        :param str color: rgb color string (default=blue)

        """

        if len(message) > 200:
            message = "..." + message[-200:]
        if color != "":
            self._lbl_status_preset.config(fg=color)
        self._status.set("  " + message)

    def clear_status(self):
        """
        Clears status bar.
        """

        self._connection.set("")
        self._status.set("")

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame

        :param event event: resize event

        """

        self.width, self.height = self.get_size()

    def get_size(self):
        """
        Get current frame size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about any resizing
        width = self.winfo_width()
        height = self.winfo_height()
        return (width, height)



"""
map_canvas.py

Multi-purpose canvas for offline and online maps.

This handles a canvas containing a location map which can be either:

 - one or more fixed offline maps based on user-provided georeferenced
   images e.g. geoTIFF (defaults to Mercator world image).
 - dynamic online map or satellite image accessed via a MapQuest API.

NOTE: The free MapQuest API key is subject to a limit of 15,000
transactions / month, or roughly 500 / day, so the map updates are only
run periodically (once a minute). This utility is NOT intended to be used for
real time navigation.

Created on 13 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

# pylint: disable=too-many-positional-arguments, too-many-arguments

from http.client import responses
from io import BytesIO
from math import sqrt
from random import randrange
from tkinter import (
    ALL,
    CENTER,
    Canvas,
    S,
)

from PIL import Image, ImageTk, UnidentifiedImageError
from pynmeagps import planar
from requests import ConnectionError as ConnError
from requests import ConnectTimeout, RequestException, get

from pygpsclient.globals import (
    BGCOL,
    CUSTOM,
    ERRCOL,
    ICON_END,
    ICON_START,
    IMG_WORLD,
    IMG_WORLD_BOUNDS,
    IMPORT,
    PNTCOL,
    WORLD,
    Area,
    AreaXY,
    Point,
)
from pygpsclient.helpers import (
    area_in_bounds,
    fontheight,
    get_track_bounds,
    ll2xy,
    normalise_area,
    point_in_bounds,
    scale_font,
)
from pygpsclient.mapquest import (
    HYB,
    MAP,
    MAPQTIMEOUT,
    POINTLIMIT,
    SAT,
    format_mapquest_request,
)
from pygpsclient.strings import (
    DLGGPXOOB,
    MAPCONFIGERR,
    MAPOPENERR,
    NOCONN,
    NOWEBMAPCONN,
    NOWEBMAPFIX,
    NOWEBMAPHTTP,
    NOWEBMAPKEY,
    OUTOFBOUNDS,
)

ZOOM = 10
POSCOL = ERRCOL
TRK_COL = "magenta"  # color of track
HACCCOL = "skyblue"
MARKERCOL = "red"
TAG_TRACK = "trak"
TAG_MARKER = "mark"
TAG_HACC = "hacc"
TAG_CLOCK = "clok"
TAG_LOCATION = "loc"
MARKERSIZE = 6
MAX_SIZE = 100000000  # 154,746,100 pixels for PIL/Image
"""Maximum image size allowed by PIL Image library"""
MAPTYPES = (WORLD, HYB, SAT, MAP, CUSTOM)
""" Map Types """


class MapCanvas(Canvas):  # pylint: disable=too-many-ancestors
    """
    Map Canvas class.
    """

    def __init__(self, app, container, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param Frame container: reference to container frame
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.__container = container  # container frame
        self.width = kwargs.get("width", 400)
        self.height = kwargs.get("height", 400)
        self._img_start = ImageTk.PhotoImage(Image.open(ICON_START))
        self._img_end = ImageTk.PhotoImage(Image.open(ICON_END))
        self._img = None
        self._mapimage = None
        self._bounds = None
        self._native_bounds = None
        self._track = None
        self._marker = None
        self._zoom = None
        self._zoommin = False
        self._last_map_update = 0
        self._last_image = None
        self._last_bounds = None
        self._lastmaptype = ""
        self._lastmappath = ""
        self._font = self.__app.font_sm
        self._fonth = fontheight(self._font)

        super().__init__(self.__container, *args, **kwargs)

        self.config(background=BGCOL)
        self.bind("<Configure>", self._on_resize)
        self.bind("<Double-Button-2>", self.on_clear)
        self.bind("<Double-Button-3>", self.on_clear)

    def draw_map(
        self,
        maptype: str = WORLD,
        location: Point = None,
        marker: Point = None,
        track: list = None,
        hacc: float = None,
        mappath: str = None,
        bounds: Area = None,
        zoom: int = ZOOM,
    ):
        """
        Draw selected map type on canvas.

        :param str maptype: map type (CUSTOM/IMPORT/MAP/SAT/WORLD)
        :param Point location: location to draw on map
        :param Point marker: marker to draw on map
        :param list track: track to draw on map
        :param float hacc: horizontal accuracy in meters
        :param str mappath: path to map image (if known)
        :param Area bounds: bounds (extent) of map image (if known)
        :param int zoom: zoom level
        """

        self._zoom = zoom
        if maptype in (WORLD, CUSTOM, IMPORT):
            self._draw_offline_map(
                maptype, location, marker, track, hacc, mappath, bounds, zoom
            )
        elif maptype in (MAP, SAT, HYB):
            self._draw_online_map(maptype, location, marker, track, hacc, bounds, zoom)

    def _draw_offline_map(
        self,
        maptype: str,
        location: Point,
        marker: Point,
        track: list,
        hacc: float,
        mappath: str = None,
        bounds: Area = None,
        zoom: int = ZOOM,
    ):  # pylint: disable=unused-argument
        """
        Draw fixed offline map using optional user-provided georeferenced
        image path(s) and calibration bounding box(es).

        Defaults to Mercator world image with bounding box [90, -180, -90, 180]
        if location is not within bounds of any custom map.

        :param str maptype: offline map type (CUSTOM/IMPORT/WORLD)
        :param Point location: location to draw on map
        :param Point marker: marker to draw on map
        :param list track: track to draw on map
        :param float hacc: horizontal accuracy in meters
        :param str mappath: path to map image (if known)
        :param Area bounds: bounds (extent) of map image (if known)
        :param int zoom: zoom level
        """

        err = self.open_offline_map(maptype, location, mappath, bounds)
        if err != "":
            self.draw_msg(err, ERRCOL)
            return

        self._lastmaptype = maptype
        self.delete(ALL)
        if (
            zoom is None
            or location is None
            or maptype == WORLD
            or self.__app.configuration.get("mapzoom_disabled_b")
        ):
            image = self._mapimage
        else:
            image, self._bounds = self._zoom_offline_map(
                self._mapimage, self._native_bounds, location, zoom
            )
            self._last_image = image
            self._last_bounds = self._bounds
        if image is None:
            self.draw_msg(DLGGPXOOB, ERRCOL)
            return
        self._img = ImageTk.PhotoImage(image.resize((self.width, self.height)))
        self.create_image(
            self.width / 2, self.height / 2, image=self._img, anchor=CENTER
        )

        if location is not None:
            self.draw_marker(location, TAG_LOCATION)
        if marker is not None:
            self.draw_marker(marker, TAG_MARKER)
        if track is not None:
            self.draw_track(track)
        if location is not None and hacc is not None:
            self.draw_hacc(location, hacc)

    def open_offline_map(
        self,
        maptype: str,
        location: Point,
        mappath: str = None,
        bounds: Area = None,
    ) -> str:
        """
        Open map image at path, or find first available map image
        from usermaps_l list which includes location.

        :param str maptype: map type (CUSTOM/IMPORT/WORLD)
        :param Point location: location
        :param str mappath: path to map image, if known
        :param Area bounds: bounds (extent) of map image, if known
        :return: error code
        :rtype: str
        """

        err = OUTOFBOUNDS.format("unknown")
        try:
            mpath = None
            if maptype == IMPORT:
                mpath = mappath
                self._bounds = bounds
            elif maptype == WORLD:
                mpath = IMG_WORLD
                self._bounds = IMG_WORLD_BOUNDS
            else:
                mpath = self._find_offline_map(location, bounds)
            if mpath is None:
                err = OUTOFBOUNDS.format("bounds" if bounds is not None else "location")
            elif self._lastmappath == mpath:  # don't bother opening again
                err = ""
            else:
                self._mapimage = Image.open(mpath)
                self._lastmappath = mpath
                err = ""
        except (ValueError, IndexError):
            err = MAPCONFIGERR
        except (FileNotFoundError, UnidentifiedImageError):
            err = MAPOPENERR.format(mpath.split("/")[-1])

        return err

    def _find_offline_map(self, location: Point, bounds: Area) -> str:
        """
        Find first map image with bounds containing location.

        :param Point location: location
        :param Area bounds: native extents of map image, if known
        :return: map path
        :rtype: str
        """
        # pylint: disable=arguments-out-of-order

        mpath = None
        usermaps = self.__app.configuration.get("usermaps_l")
        for fpath, extent in usermaps:
            extents = normalise_area((extent[0], extent[1], extent[2], extent[3]))
            # check if bounds or location are within map extents
            if (bounds is not None and area_in_bounds(extents, bounds)) or (
                location is not None and point_in_bounds(extents, location)
            ):
                mpath = fpath
                self._bounds = extents
                self._native_bounds = extents
                break
        return mpath

    def _zoom_offline_map(
        self, image: Image, extents: Area, location: Point, zoom: int
    ) -> tuple:
        """
        Zoom (crop) offline image centered at location and
        calculate new bounding box.

        Automatically increments zoom until image is
        entirely within zoom bounds.

        ;param Image image: native map image
        :param Area extents: native map extents
        :param Point location: location (center point)
        :param int zoom: zoom level
        :return: tuple of (zoomed image, zoomed bounds)
        :rtype: tuple
        """

        zoombounds = self.zoom_bounds(self.height, self.width, location, zoom, CUSTOM)
        self._zoom = zoom
        (x1, y1), (x2, y2) = [
            ll2xy(image.width, image.height, extents, pnt)
            for pnt in (
                Point(zoombounds.lat2, zoombounds.lon1),
                Point(zoombounds.lat1, zoombounds.lon2),
            )
        ]

        # check if cropped image exceeds PIL's max permissible size
        size = (x2 - x1) * (y2 - y1)
        if size >= MAX_SIZE:
            self._zoommin = True
            return self._last_image, self._last_bounds

        self._zoommin = False
        return image.crop(AreaXY(x1, y1, x2, y2)), zoombounds

    def _draw_online_map(
        self,
        maptype: str,
        location: Point,
        marker: Point,
        track: list,
        hacc: float,
        bounds: Area = None,
        zoom: int = ZOOM,
    ):  # pylint: disable=unused-argument
        """
        Draw scalable web map or satellite image via online MapQuest API.

        :param str maptype: online map type (MAP/SAT)
        :param Point location: location to draw on map
        :param Point marker: marker to draw on map
        :param list track: track to draw on map
        :param float hacc: horizontal accuracy in meters
        :param Area bounds: bounds (extent) of map image, if known
        :param int zoom: zoom level
        """

        sc = NOCONN
        err = ""
        hacc = hacc if isinstance(hacc, (float, int)) else 0

        if maptype != self._lastmaptype:
            self._lastmaptype = maptype

        mqapikey = self.__app.configuration.get("mqapikey_s")
        if mqapikey == "":
            self.draw_msg(NOWEBMAPKEY, ERRCOL)
            return

        if track is not None:
            points = track
        elif location is not None:
            points = [
                location,
            ]
        else:
            self.draw_msg(NOWEBMAPFIX, ERRCOL)
            return

        if bounds is None:
            # set bounds relative to location and zoom
            if zoom > 1:
                bounds = self.zoom_bounds(
                    self.height, self.width, location, zoom, maptype
                )
        url = format_mapquest_request(
            mqapikey,
            maptype,
            self.width,
            self.height,
            zoom,
            points,  # list
            bounds,  # bbox
            hacc,
        )

        try:
            response = get(url, timeout=MAPQTIMEOUT)
            sc = responses[response.status_code]  # get descriptive HTTP status
            response.raise_for_status()  # raise Exception on HTTP error
            if sc == "OK":
                self._bounds = bounds
                img_data = response.content
                self._img = ImageTk.PhotoImage(Image.open(BytesIO(img_data)))
                self.delete(ALL)
                self.create_image(
                    self.width / 2, self.height / 2, image=self._img, anchor=CENTER
                )
                self.update_idletasks()
                return
        except (ConnError, ConnectTimeout):
            err = NOWEBMAPCONN
        except RequestException:
            err = NOWEBMAPHTTP.format(sc)

        self.draw_msg(err, ERRCOL)

    def draw_track(self, track: list):
        """
        Draw track on canvas.

        :param list track: list of track points
        """

        self.delete(TAG_TRACK)
        i = 0
        for i, pnt in enumerate(track):
            x, y = ll2xy(self.width, self.height, self._bounds, Point(pnt.lat, pnt.lon))
            if i:
                x2, y2 = x, y
                self.create_line(x1, y1, x2, y2, fill=TRK_COL, width=3, tags=TAG_TRACK)
                x1, y1 = x2, y2
            else:
                x1, y1 = x, y
                xstart, ystart = x, y
        if i:
            self.create_image(
                xstart, ystart, image=self._img_start, anchor=S, tags=TAG_TRACK
            )
            self.create_image(x2, y2, image=self._img_end, anchor=S, tags=TAG_TRACK)

    def draw_marker(self, marker: Point, markertype: str = TAG_LOCATION):
        """
        Draw marker point on canvas

        :param Point marker: marker point
        :param str markertype: marker type
        """

        x, y = ll2xy(self.width, self.height, self._bounds, marker)
        if markertype == TAG_MARKER:
            self.delete(TAG_MARKER)
            self.create_circle(
                x, y, 2, outline=MARKERCOL, fill=MARKERCOL, tags=TAG_MARKER
            )
            self.create_text(
                x,
                y,
                text=f"{marker.lat:.08f}\n{marker.lon:.08f}",
                anchor=CENTER,
                fill=MARKERCOL,
                tags=TAG_MARKER,
            )
        else:
            self.delete(TAG_LOCATION)
            self.create_line(
                x,
                y - MARKERSIZE,
                x,
                y + MARKERSIZE,
                fill=MARKERCOL,
                width=2,
                tags=TAG_LOCATION,
            )
            self.create_line(
                x - MARKERSIZE,
                y,
                x + MARKERSIZE,
                y,
                fill=MARKERCOL,
                width=2,
                tags=TAG_LOCATION,
            )

    def draw_hacc(self, location: Point, hacc: float):
        """
        Draw horizontal accuracy perimeter on canvas.

        :param float hacc: horizontal accurancy in meters
        """

        # FYI not possible to draw translucent circles in tkinter
        # other than by messing around with stippled polygons
        self.delete(TAG_HACC)
        radius = int(
            sqrt(self.height**2 + self.width**2)
            * hacc
            / planar(
                self._bounds.lat1,
                self._bounds.lon1,
                self._bounds.lat2,
                self._bounds.lon2,
            )
        )
        x, y = ll2xy(self.width, self.height, self._bounds, location)
        self.create_circle(x, y, radius, outline=HACCCOL, fill="", tags=TAG_HACC)

    def draw_countdown(self, wait: int):
        """
        Draw clock icon indicating time until next scheduled map refresh.

        :param int wait: wait time in seconds
        """

        self.delete(TAG_CLOCK)
        self.create_oval((5, 5, 20, 20), fill="", outline="black", tags=TAG_CLOCK)
        self.create_arc(
            (5, 5, 20, 20),
            start=90,
            extent=wait,
            fill="black",
            outline="black",
            tags=TAG_CLOCK,
        )

    def draw_msg(self, msg: str, color: str = PNTCOL):
        """
        Draw message on canvas.

        :param str msg: message
        :param str color: color
        """

        w, h = self.get_size()
        self.delete(ALL)
        self.create_text(
            w / 2,
            h / 2,
            text=msg,
            fill=color,
            font=self._font,
            anchor=CENTER,
        )

    def on_clear(self, event):  # pylint: disable=unused-argument
        """
        Clear map image.
        """

        self.delete(ALL)
        self._marker = None

    def _on_resize(self, event):  # pylint: disable=unused-argument
        """
        Resize frame

        :param event event: resize event
        """

        self.width, self.height = self.get_size()
        self._font, self._fonth = scale_font(self.width, 10, 25, 20)

    def get_size(self) -> tuple:
        """
        Get current canvas size.

        :return: window size (width, height)
        :rtype: tuple
        """

        self.update_idletasks()  # Make sure we know about any resizing
        return self.winfo_width(), self.winfo_height()

    @staticmethod
    def zoom_bounds(
        height: int, width: int, location: Point, zoom: int, maptype: str = CUSTOM
    ) -> Area:
        """
        Get map bounds for given zoom level.

        :param int height: height
        :param int width: width
        :param Point location: location
        :param int zoom: zoom level
        :param str maptype: map type
        :return: zoom bounds
        :rtype: Area
        """

        if location is None:
            return None
        xoff = 90 / 2**zoom
        yoff = xoff * height / width
        return Area(
            location.lat - yoff,
            location.lon - xoff,
            location.lat + yoff,
            location.lon + xoff,
        )

    @property
    def bounds(self) -> Area:
        """
        Getter for custom map bounds.

        :return: bounds of displayed map
        :rtype: Area or None
        """

        return self._bounds

    @property
    def marker(self) -> Point:
        """
        Getter for marker.

        :return: marked location
        :rtype: Point
        """

        return self._marker

    @property
    def track(self) -> list:
        """
        Getter for track.

        :return: recorded track
        :rtype: list or None
        """

        return self._track

    @property
    def trackbounds(self) -> tuple:
        """
        Getter for track bounds and center.

        :return: tuple of (bounds, center point)
        :rtype: (Area, Point)
        """

        if isinstance(self._track, list):
            if len(self._track) > 0:
                return get_track_bounds(self._track)
        return None, None

    @property
    def zoom(self) -> int:
        """
        Getter for zoom.

        :return: applied zoom
        :rtype: int or None
        """

        return self._zoom

    @property
    def zoommin(self) -> bool:
        """
        Getter for zoommin flag.

        :return: if max zoom reached
        :rtype: bool
        """

        return self._zoommin

    @track.setter
    def track(self, location: Point):
        """
        Update or reset track list.
        Set location to None to reset.

        :param Point location: location
        """

        if location is None:
            self._track = None
            return
        if self._track is None:
            self._track = []

        if len(self._track) > 0:  # only record if different from previous
            if round(self._track[-1][0], 8) != round(location.lat, 8) or round(
                self._track[-1][1], 8
            ) != round(location.lon, 8):
                self._track.append(Point(location.lat, location.lon))
        else:
            self._track.append(Point(location.lat, location.lon))
        # limit size of track list
        while len(self._track) > POINTLIMIT:
            self._track.pop(randrange(1, len(self._track) - 1))



"""
ubx_cfgval_frame.py

UBX Configuration frame for CFG-VAL commands.

Created on 22 Dec 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from tkinter import (
    HORIZONTAL,
    LEFT,
    NORMAL,
    VERTICAL,
    Button,
    E,
    Entry,
    Frame,
    IntVar,
    Label,
    Listbox,
    N,
    Radiobutton,
    S,
    Scrollbar,
    Spinbox,
    StringVar,
    W,
)

from PIL import Image, ImageTk
from pyubx2 import UBX_CONFIG_DATABASE, UBXMessage
from pyubx2.ubxhelpers import attsiz, atttyp, cfgname2key

from pygpsclient.globals import (
    ERRCOL,
    ICON_CONFIRMED,
    ICON_PENDING,
    ICON_SEND,
    ICON_WARNING,
    OKCOL,
    READONLY,
    UBX_CFGVAL,
)

VALSET = 0
VALDEL = 1
VALGET = 2
ATTDICT = {
    "U": "unsigned int",
    "I": "signed int",
    "R": "float",
    "L": "bool",
    "E": "unsigned int",
    "X": "byte(s) as hex string",
    "C": "char(s)",
}


class UBX_CFGVAL_Frame(Frame):
    """
    UBX CFG-VAL configuration command panel.
    """

    def __init__(self, app, container, *args, **kwargs):
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param Frame container: reference to container frame (config-dialog)
        :param args: optional args to pass to Frame parent class
        :param kwargs: optional kwargs to pass to Frame parent class
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.__container = container

        super().__init__(container.container, *args, **kwargs)

        self._img_send = ImageTk.PhotoImage(Image.open(ICON_SEND))
        self._img_pending = ImageTk.PhotoImage(Image.open(ICON_PENDING))
        self._img_confirmed = ImageTk.PhotoImage(Image.open(ICON_CONFIRMED))
        self._img_warn = ImageTk.PhotoImage(Image.open(ICON_WARNING))
        self._cfgval_cat = None
        self._cfgval_keyname = None
        self._cfgval_keyid = None
        self._cfgmode = IntVar()
        self._cfgatt = StringVar()
        self._cfgkeyid = StringVar()
        self._cfgval = StringVar()
        self._cfglayer = StringVar()

        self._body()
        self._do_layout()
        self._attach_events()
        self.reset()

    def _body(self):
        """
        Set up frame and widgets.
        """

        self._lbl_configdb = Label(
            self, text="CFG-VALSET/DEL/GET Configuration Interface", anchor=W
        )
        self._lbl_cat = Label(self, text="Category", anchor=W)
        self._lbx_cat = Listbox(
            self,
            border=2,
            relief="sunken",
            height=12,
            justify=LEFT,
            exportselection=False,
        )
        self._scr_catv = Scrollbar(self, orient=VERTICAL)
        self._scr_cath = Scrollbar(self, orient=HORIZONTAL)
        self._lbx_cat.config(yscrollcommand=self._scr_catv.set)
        self._lbx_cat.config(xscrollcommand=self._scr_cath.set)
        self._scr_catv.config(command=self._lbx_cat.yview)
        self._scr_cath.config(command=self._lbx_cat.xview)
        self._lbl_parm = Label(self, text="Keyname", anchor=W)
        self._lbx_parm = Listbox(
            self,
            border=2,
            relief="sunken",
            height=12,
            justify=LEFT,
            exportselection=False,
        )
        self._scr_parmv = Scrollbar(self, orient=VERTICAL)
        self._scr_parmh = Scrollbar(self, orient=HORIZONTAL)
        self._lbx_parm.config(yscrollcommand=self._scr_parmv.set)
        self._lbx_parm.config(xscrollcommand=self._scr_parmh.set)
        self._scr_parmv.config(command=self._lbx_parm.yview)
        self._scr_parmh.config(command=self._lbx_parm.xview)

        self._rad_cfgset = Radiobutton(
            self, text="CFG-VALSET", variable=self._cfgmode, value=0
        )
        self._rad_cfgdel = Radiobutton(
            self, text="CFG-VALDEL", variable=self._cfgmode, value=1
        )
        self._rad_cfgget = Radiobutton(
            self, text="CFG-VALGET", variable=self._cfgmode, value=2
        )
        self._lbl_key = Label(self, text="KeyID")
        self._lbl_keyid = Label(
            self,
            textvariable=self._cfgkeyid,
            width=10,
            border=1,
            relief="sunken",
        )
        self._lbl_type = Label(self, text="Type")
        self._lbl_att = Label(
            self,
            textvariable=self._cfgatt,
            width=5,
            border=1,
            relief="sunken",
        )
        self._lbl_layer = Label(self, text="Layer")
        self._spn_layer = Spinbox(
            self,
            textvariable=self._cfglayer,
            values=("RAM", "BBR", "FLASH", "DEFAULT"),
            wrap=True,
            width=8,
            state=READONLY,
        )
        self._lbl_val = Label(self, text="Value")
        self._ent_val = Entry(
            self,
            textvariable=self._cfgval,
            state=READONLY,
            relief="sunken",
        )

        self._lbl_send_command = Label(self)
        self._btn_send_command = Button(
            self,
            image=self._img_send,
            width=50,
            command=self._on_send_config,
        )

    def _do_layout(self):
        """
        Layout widgets.
        """

        self._lbl_configdb.grid(column=0, row=0, columnspan=6, padx=3, sticky=(W, E))
        self._lbl_cat.grid(column=0, row=1, padx=3, sticky=(W, E))
        self._lbx_cat.grid(column=0, row=2, rowspan=5, padx=3, pady=3, sticky=(W, E))
        self._scr_catv.grid(column=0, row=2, rowspan=5, sticky=(N, S, E))
        self._scr_cath.grid(column=0, row=7, sticky=(W, E))
        self._lbl_parm.grid(column=1, row=1, columnspan=4, padx=3, sticky=(W, E))
        self._lbx_parm.grid(
            column=1, row=2, columnspan=4, rowspan=5, padx=3, pady=3, sticky=(W, E)
        )
        self._scr_parmv.grid(column=4, row=2, rowspan=5, sticky=(N, S, E))
        self._scr_parmh.grid(column=1, row=7, columnspan=4, sticky=(W, E))
        self._rad_cfgget.grid(column=0, row=8, padx=3, pady=0, sticky=W)
        self._rad_cfgset.grid(column=0, row=9, padx=3, pady=0, sticky=W)
        self._rad_cfgdel.grid(column=0, row=10, padx=3, pady=0, sticky=W)
        self._lbl_key.grid(column=1, row=8, padx=3, pady=0, sticky=E)
        self._lbl_keyid.grid(column=2, row=8, padx=3, pady=0, sticky=W)
        self._lbl_type.grid(column=3, row=8, padx=3, pady=0, sticky=E)
        self._lbl_att.grid(column=4, row=8, padx=3, pady=0, sticky=W)
        self._lbl_layer.grid(column=1, row=9, padx=3, pady=0, sticky=E)
        self._spn_layer.grid(column=2, row=9, padx=3, pady=0, sticky=W)
        self._lbl_val.grid(column=1, row=10, padx=3, pady=0, sticky=E)
        self._ent_val.grid(
            column=2, row=10, columnspan=3, padx=3, pady=0, sticky=(W, E)
        )

        self._btn_send_command.grid(
            column=3, row=12, rowspan=2, ipadx=3, ipady=3, sticky=E
        )
        self._lbl_send_command.grid(
            column=4, row=13, rowspan=2, ipadx=3, ipady=3, sticky=E
        )

        (cols, rows) = self.grid_size()
        for i in range(cols):
            self.grid_columnconfigure(i, weight=1)
        for i in range(rows):
            self.grid_rowconfigure(i, weight=1)
        self.option_add("*Font", self.__app.font_sm)

    def _attach_events(self):
        """
        Bind events to widgets.
        """

        self._lbx_cat.bind("<<ListboxSelect>>", self._on_select_cat)
        self._lbx_parm.bind("<<ListboxSelect>>", self._on_select_parm)
        self._cfgmode.trace_add("write", self._on_select_mode)

    def reset(self):
        """
        Reset panel with sorted list of unique UBX Config Database key categories.
        """

        cdb_cats = []
        for cdb in UBX_CONFIG_DATABASE:
            cdbs = cdb.split("_", maxsplit=3)
            cdbp = f"{cdbs[0]}_{cdbs[1]}"
            if cdbs[1] == "MSGOUT":  # subdivide large MSGOUT category
                cdbp += f"_{cdbs[2]}"
            if cdbp not in cdb_cats:
                cdb_cats.append(cdbp)

        for i, cat in enumerate(cdb_cats):
            self._lbx_cat.insert(i, cat)
        self._cfgmode.set(2)

    def _on_select_mode(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Mode has been selected.
        """

        if self._cfgmode.get() == VALSET:
            self._ent_val.config(state=NORMAL)
        else:
            self._ent_val.config(state=READONLY)

    def _on_select_cat(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Configuration category has been selected.
        """

        idx = self._lbx_cat.curselection()
        self._cfgval_cat = self._lbx_cat.get(idx)

        self._lbx_parm.delete(0, "end")
        self._cfgkeyid.set("")
        self._cfgatt.set("")
        idx = 0
        for keyname, (_, _) in UBX_CONFIG_DATABASE.items():
            if self._cfgval_cat in keyname:
                self._lbx_parm.insert(idx, keyname)
                idx += 1
        self._cfgval.set("")

    def _on_select_parm(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Configuration parameter (keyname) has been selected.
        """

        idx = self._lbx_parm.curselection()
        self._cfgval_keyname = self._lbx_parm.get(idx)

        (keyid, att) = cfgname2key(self._cfgval_keyname)
        self._cfgkeyid.set(hex(keyid))
        self._cfgatt.set(att)
        self._cfgval.set("")

    def _on_send_config(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Config interface send button has been clicked.
        """

        if self._cfgval_keyname is not None:
            if self._cfgmode.get() == VALSET:
                self._do_valset()
            elif self._cfgmode.get() == VALDEL:
                self._do_valdel()
            else:
                self._do_valget()

    def _do_valset(self):
        """
        Send a CFG-VALSET message.

        :return: valid entry flag
        :rtype: bool
        """

        valid_entry = True
        att = atttyp(self._cfgatt.get())
        atts = attsiz(self._cfgatt.get())
        val = self._cfgval.get()
        layers = self._cfglayer.get()
        if layers == "BBR":
            layers = 2
        elif layers == "FLASH":
            layers = 4
        else:
            layers = 1
        try:
            if att in ("C", "X"):  # byte or char
                if len(val) == atts * 2:  # 2 hex chars per byte
                    val = bytearray.fromhex(val)
                else:
                    valid_entry = False
            elif att in ("E", "U"):  # unsigned integer
                val = int(val)
                if val < 0:
                    valid_entry = False
            elif att == "L":  # bool
                val = int(val)
                if val not in (0, 1):
                    valid_entry = False
            elif att == "I":  # signed integer
                val = int(val)
            elif att == "R":  # floating point
                val = float(val)
            transaction = 0
            cfgData = [
                (self._cfgval_keyname, val),
            ]
        except ValueError:
            valid_entry = False

        if valid_entry:
            msg = UBXMessage.config_set(layers, transaction, cfgData)
            self.__container.send_command(msg)
            self._lbl_send_command.config(image=self._img_pending)
            self.__container.set_status(
                "CFG-VALSET SET message sent",
            )
            for msgid in ("ACK-ACK", "ACK-NAK"):
                self.__container.set_pending(msgid, UBX_CFGVAL)
        else:
            self._lbl_send_command.config(image=self._img_warn)
            typ = ATTDICT[att]
            self.__container.set_status(
                (
                    "INVALID ENTRY - must conform to parameter "
                    f"type {att} ({typ}) and size {atts} bytes"
                ),
                ERRCOL,
            )

        return valid_entry

    def _do_valdel(self):
        """
        Send a CFG-VALDEL message.
        """

        layers = self._cfglayer.get()
        if layers == "BBR":
            layers = 2
        elif layers == "FLASH":
            layers = 4
        else:
            layers = 1
        transaction = 0
        key = [
            self._cfgval_keyname,
        ]
        msg = UBXMessage.config_del(layers, transaction, key)
        self.__container.send_command(msg)
        self._lbl_send_command.config(image=self._img_pending)
        self.__container.set_status("CFG-VALDEL SET message sent")
        for msgid in ("ACK-ACK", "ACK-NAK"):
            self.__container.set_pending(msgid, UBX_CFGVAL)

    def _do_valget(self):
        """
        Send a CFG-VALGET message.
        """

        layers = self._cfglayer.get()
        if layers == "BBR":
            layers = 1
        elif layers == "FLASH":
            layers = 2
        elif layers == "DEFAULT":
            layers = 7
        else:
            layers = 0
        transaction = 0
        keys = [
            self._cfgval_keyname,
        ]
        msg = UBXMessage.config_poll(layers, transaction, keys)
        self.__container.send_command(msg)
        self._lbl_send_command.config(image=self._img_pending)
        self.__container.set_status("CFG-VALGET POLL message sent")
        for msgid in ("CFG-VALGET", "ACK-ACK", "ACK-NAK"):
            self.__container.set_pending(msgid, UBX_CFGVAL)

    def update_status(self, msg: UBXMessage):  # pylint: disable=unused-argument
        """
        Update pending confirmation status.

        :param UBXMessage msg: UBX config message
        """

        if msg.identity == "CFG-VALGET":
            self._lbl_send_command.config(image=self._img_confirmed)
            val = getattr(msg, self._cfgval_keyname, None)
            if val is not None:
                if isinstance(val, bytes):
                    val = val.hex()
                self._cfgval.set(val)
            self.__container.set_status("CFG-VALGET GET message received", OKCOL)

        elif msg.identity == "ACK-ACK":
            self._lbl_send_command.config(image=self._img_confirmed)
            self.__container.set_status("CFG-VAL command acknowledged", OKCOL)

        elif msg.identity == "ACK-NAK":
            self._lbl_send_command.config(image=self._img_warn)
            self.__container.set_status("CFG-VAL command rejected", ERRCOL)



"""
about_dialog.py

About Dialog Box class for PyGPSClient application.

Created on 20 Sep 2020

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

import logging
from inspect import currentframe, getfile
from os import path
from platform import python_version
from subprocess import CalledProcessError, run
from sys import executable
from tkinter import Button, Checkbutton, E, Frame, IntVar, Label, Tcl, W
from webbrowser import open_new_tab

from PIL import Image, ImageTk
from pygnssutils import version as PGVERSION
from pynmeagps import version as NMEAVERSION
from pyrtcm import version as RTCMVERSION
from pysbf2 import version as SBFVERSION
from pyspartn import version as SPARTNVERSION
from pyubx2 import version as UBXVERSION

from pygpsclient._version import __version__ as VERSION
from pygpsclient.globals import (
    ERRCOL,
    ICON_APP128,
    ICON_GITHUB,
    ICON_SPONSOR,
    INFOCOL,
    LICENSE_URL,
    OKCOL,
    SPONSOR_URL,
)
from pygpsclient.helpers import check_latest
from pygpsclient.strings import ABOUTTXT, COPYRIGHTTXT, DLGTABOUT, GITHUB_URL
from pygpsclient.toplevel_dialog import ToplevelDialog

LIBVERSIONS = {
    "PyGPSClient": VERSION,
    "pygnssutils": PGVERSION,
    "pyubx2": UBXVERSION,
    "pysbf2": SBFVERSION,
    "pynmeagps": NMEAVERSION,
    "pyrtcm": RTCMVERSION,
    "pyspartn": SPARTNVERSION,
}


MINDIM = (600, 400)


class AboutDialog(ToplevelDialog):
    """
    About dialog box class
    """

    def __init__(self, app, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Initialise Toplevel dialog

        :param Frame app: reference to main tkinter application
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.logger = logging.getLogger(__name__)
        self._img_icon = ImageTk.PhotoImage(Image.open(ICON_APP128).resize((64, 64)))
        self._img_github = ImageTk.PhotoImage(Image.open(ICON_GITHUB).resize((32, 32)))
        self._img_sponsor = ImageTk.PhotoImage(Image.open(ICON_SPONSOR))
        self._checkonstartup = IntVar()
        self._checkonstartup.set(self.__app.configuration.get("checkforupdate_b"))
        self._updates = []

        super().__init__(app, DLGTABOUT, MINDIM)

        self._body()
        self._do_layout()
        self._attach_events()
        self._finalise()

    def _body(self):
        """
        Set up widgets.
        """

        self._frm_body = Frame(self.container)
        self._lbl_icon = Label(self._frm_body, image=self._img_icon, borderwidth=0)
        self._lbl_desc = Label(
            self._frm_body,
            text=ABOUTTXT,
            wraplength=350,
            borderwidth=0,
        )
        tkv = Tcl().call("info", "patchlevel")
        self._lbl_python_version = Label(
            self._frm_body,
            text=f"Python: {python_version()}  Tk: {tkv}  Spatial: {self.__app.db_enabled}",
        )
        self._lbl_lib_versions = []
        for nam, ver in LIBVERSIONS.items():
            self._lbl_lib_versions.append(
                Label(
                    self._frm_body,
                    text=f"{nam}: {ver}",
                    borderwidth=0,
                    highlightthickness=0,
                )
            )
        self._btn_checkupdate = Button(
            self._frm_body,
            text="Check for updates",
            width=12,
            cursor="hand2",
        )
        self._chk_checkupdate = Checkbutton(
            self._frm_body,
            text="Check on startup",
            variable=self._checkonstartup,
        )
        self._lbl_giticon = Label(
            self._frm_body,
            image=self._img_github,
            cursor="hand2",
        )
        self._lbl_sponsoricon = Label(
            self._frm_body,
            image=self._img_sponsor,
            cursor="hand2",
        )
        self._lbl_github = Label(
            self._frm_body,
            text=GITHUB_URL,
            fg=INFOCOL,
            cursor="hand2",
        )
        self._lbl_copyright = Label(
            self._frm_body,
            text=COPYRIGHTTXT,
            cursor="hand2",
        )

    def _do_layout(self):
        """
        Arrange widgets in dialog.
        """

        self._frm_body.grid(column=0, row=0, padx=5, pady=5, ipadx=5, ipady=5)
        self._lbl_icon.grid(column=0, row=1, columnspan=2, padx=3, pady=0)
        self._lbl_desc.grid(column=0, row=2, columnspan=2, padx=3, pady=0)
        self._lbl_python_version.grid(column=0, row=3, columnspan=2, padx=3, pady=1)
        i = 0
        for i, _ in enumerate(LIBVERSIONS):
            self._lbl_lib_versions[i].grid(column=0, row=4 + i, columnspan=2, padx=2)
        self._btn_checkupdate.grid(
            column=0, row=5 + i, ipadx=3, ipady=3, padx=3, pady=3
        )
        self._chk_checkupdate.grid(
            column=1, row=5 + i, ipadx=3, ipady=3, padx=3, pady=3
        )
        self._lbl_giticon.grid(column=0, row=6 + i, padx=(3, 1), pady=3, sticky=E)
        self._lbl_sponsoricon.grid(column=1, row=6 + i, padx=(3, 1), pady=3, sticky=W)
        self._lbl_github.grid(column=0, row=7 + i, columnspan=2, padx=(1, 3), pady=3)
        self._lbl_copyright.grid(column=0, row=8 + i, columnspan=2, padx=3, pady=3)

    def _attach_events(self):
        """
        Bind events to dialog.
        """

        self._btn_checkupdate.bind("<Button>", self._check_for_update)
        self._lbl_giticon.bind("<Button>", self._on_github)
        self._lbl_github.bind("<Button>", self._on_github)
        self._lbl_sponsoricon.bind("<Button>", self._on_sponsor)
        self._lbl_copyright.bind("<Button>", self._on_license)
        self._btn_exit.focus_set()
        self._checkonstartup.trace_add("write", self._on_update_config)

    def _on_update_config(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Save current settings to saved app config dict.
        """

        self.__app.configuration.set(
            "checkforupdate_b", int(self._checkonstartup.get())
        )

    def _on_github(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Close dialog and go to GitHub.
        """

        open_new_tab(GITHUB_URL)
        self.on_exit()

    def _on_sponsor(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Close dialog and go to Sponsor website.
        """

        open_new_tab(SPONSOR_URL)
        self.on_exit()

    def _on_license(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Close dialog and go to GitHub LICENSE file.
        """

        open_new_tab(LICENSE_URL)
        self.on_exit()

    def _check_for_update(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Check for updates.
        """

        self._updates = []
        for i, (nam, current) in enumerate(LIBVERSIONS.items()):
            latest = check_latest(nam)
            txt = f"{nam}: {current}"
            if latest == current:
                txt += " ✓"
                col = OKCOL
            elif latest == "N/A":
                txt += " - Info not available!"
                col = ERRCOL
            else:
                self._updates.append(nam)
                txt += f" - Latest version is {latest}"
                col = ERRCOL
            self._lbl_lib_versions[i].config(text=txt, fg=col)

        if len(self._updates) > 0:
            self._btn_checkupdate.config(text="UPDATE", fg=INFOCOL)
            self._btn_checkupdate.bind("<Button>", self._do_update)

    def _do_update(self, *args, **kwargs):  # pylint: disable=unused-argument
        """
        Run python update.
        """

        self._btn_checkupdate.config(text="UPDATING...", fg=INFOCOL)
        self.update_idletasks()
        pth = path.dirname(path.abspath(getfile(currentframe())))
        if "pipx" in pth:  # installed into venv using pipx
            cmd = [
                "pipx",
                "upgrade",
                "pygpsclient",
            ]
        else:  # installed using pip
            cmd = [
                executable,  # i.e. python3 or python
                "-m",
                "pip",
                "install",
                "--upgrade",
            ]
            for pkg in self._updates:
                cmd.append(pkg)

        result = None
        try:
            result = run(cmd, check=True, capture_output=True)
            self.logger.debug(result.stdout)
        except CalledProcessError:
            self.logger.error(result.stdout)
            self._btn_checkupdate.config(text="UPDATE FAILED", fg=ERRCOL)
            self._btn_checkupdate.bind("<Button>", self._check_for_update)
            return

        self._btn_checkupdate.config(text="RESTART APP", fg=OKCOL)
        self._btn_checkupdate.bind("<Button>", self.__app.on_exit)



"""
spartn_gnss_frame.py

SPARTN GNSS Receiver configuration dialog.

This handles the configuration of the GNSS receiver to receive and decrypt
incoming SPARTN RTK messages.

FYI CFG_NAVSPG_CONSTR_DGNSSTO governs how quickly rcvr drops out of RTK mode
after loss of correction signal - default is 60 seconds but this may need
to be increased if the correction source is intermittent.

Created on 26 Jan 2023

:author: semuadmin (Steve Smith)
:copyright: 2020 semuadmin
:license: BSD 3-Clause
"""

from datetime import datetime, timedelta, timezone
from tkinter import (
    NORMAL,
    Button,
    Checkbutton,
    E,
    Entry,
    Frame,
    IntVar,
    Label,
    Radiobutton,
    Spinbox,
    StringVar,
    W,
    ttk,
)

from PIL import Image, ImageTk
from pyspartn import date2timetag
from pyubx2 import POLL, SET, SET_LAYER_RAM, TXN_NONE, U1, U2, U4, UBXMessage, val2bytes

from pygpsclient.globals import (
    CONNECTED,
    ERRCOL,
    ICON_BLANK,
    ICON_CONFIRMED,
    ICON_DISCONN,
    ICON_LOAD,
    ICON_PENDING,
    ICON_SEND,
    ICON_WARNING,
    OKCOL,
    READONLY,
    RXMMSG,
    SPARTN_GNSS,
    SPARTN_KEYLEN,
    SPARTN_SOURCE_IP,
    SPARTN_SOURCE_LB,
)
from pygpsclient.helpers import (
    VALDMY,
    VALHEX,
    VALLEN,
    date2wnotow,
    parse_rxmspartnkey,
    valid_entry,
)
from pygpsclient.spartn_json_config import SpartnJsonConfig
from pygpsclient.strings import (
    CONFIGBAD,
    CONFIGOK,
    CONFIGRXM,
    DLGJSONERR,
    DLGJSONOK,
    LBLDISNMEA,
    LBLJSONLOAD,
    LBLSPARTNGN,
    LBLSPTNCURR,
    LBLSPTNDAT,
    LBLSPTNFP,
    LBLSPTNKEY,
    LBLSPTNNEXT,
    LBLSPTNUPLOAD,
    NOTCONN,
    NULLSEND,
)

U2MAX = 2e16 - 1
U8MAX = 2e64 - 1
BAUDRATE = 38400
RESERVED0 = b"\x00\x00"
RESERVED1 = b"\x00"
CFGSET = "CFG-VALGET/SET"
CFGPOLL = "CFG-VALGET"
INPORTS = ("I2C", "UART1", "UART2", "USB", "SPI")
RTK_FLOAT = 2
RTK_FIXED = 3


class SPARTNGNSSDialog(Frame):
    """,
    SPARTNConfigDialog class.
    """

    def __init__(
        self, app, container, *args, **kwargs
    ):  # pylint: disable=unused-argument
        """
        Constructor.

        :param Frame app: reference to main tkinter application
        :param Frame container: reference to container frame
        :param args: optional args to pass to parent class (not currently used)
        :param kwargs: optional kwargs to pass to parent class (not currently used)
        """

        self.__app = app  # Reference to main application class
        self.__master = self.__app.appmaster  # Reference to root class (Tk)
        self.__container = container  # container frame

        Frame.__init__(self, self.__container.container, *args, **kwargs)

        self._img_blank = ImageTk.PhotoImage(Image.open(ICON_BLANK))
        self._img_pending = ImageTk.PhotoImage(Image.open(ICON_PENDING))
        self._img_confirmed = ImageTk.PhotoImage(Image.open(ICON_CONFIRMED))
        self._img_warn = ImageTk.PhotoImage(Image.open(ICON_WARNING))
        self._img_send = ImageTk.PhotoImage(Image.open(ICON_SEND))
        self._img_disconn = ImageTk.PhotoImage(Image.open(ICON_DISCONN))
        self._img_load = ImageTk.PhotoImage(Image.open(ICON_LOAD))
        self._status = StringVar()
        self._spartn_source = IntVar()
        self._status_cfgmsg = StringVar()
        self._spartn_key1 = StringVar()
        self._spartn_valdate1 = StringVar()
        self._spartn_key2 = StringVar()
        self._spartn_valdate2 = StringVar()
        self._upload_keys = IntVar()
        self._send_f9p_config = IntVar()
        self._dgnssto = IntVar()
        self._disable_nmea = IntVar()
        self._ports = INPORTS

        self._body()
        self._do_layout()
        self._reset()
        self._attach_events()

    def _body(self):
        """
        Set up frame and widgets.
        """
        # pylint: disable=unnecessary-lambda

        # GNSS Receiver (F9P) configuration widgets
        self._lbl_gnss = Label(self, text=LBLSPARTNGN)
        self._lbl_curr = Label(self, text=LBLSPTNCURR)
        self._lbl_key1 = Label(self, text=LBLSPTNKEY)
        self._ent_key1 = Entry(
            self,
            textvariable=self._spartn_key1,
            state=NORMAL,
            relief="sunken",
            width=32,
        )
        self._lbl_valdate1 = Label(self, text=LBLSPTNDAT)
        self._ent_valdate1 = Entry(
            self,
            textvariable=self._spartn_valdate1,
            state=NORMAL,
            relief="sunken",
            width=8,
        )
        self._lbl_next = Label(self, text=LBLSPTNNEXT)
        self._lbl_key2 = Label(self, text=LBLSPTNKEY)
        self._ent_key2 = Entry(
            self,
            textvariable=self._spartn_key2,
            state=NORMAL,
            relief="sunken",
            width=32,
        )
        self._lbl_valdate2 = Label(self, text=LBLSPTNDAT)
        self._ent_valdate2 = Entry(
            self,
            textvariable=self._spartn_valdate2,
            state=NORMAL,
            relief="sunken",
            width=8,
        )
        self._rad_source0 = Radiobutton(
            self, text="IP", variable=self._spartn_source, value=SPARTN_SOURCE_IP
        )
        self._rad_source1 = Radiobutton(
            self,
            text="L-Band",
            variable=self._spartn_source,
            value=SPARTN_SOURCE_LB,
        )
        self._lbl_dgnssto = Label(
            self,
            text="DGPS Timeout",
        )
        self._spn_dgnssto = Spinbox(
            self,
            values=(60, 120, 240),
            width=4,
            state=READONLY,
            wrap=True,
            textvariable=self._dgnssto,
        )
        self._lbl_loadjson = Label(
            self,
            text=LBLJSONLOAD,
        )
        self._btn_loadjson = Button(
            self,
            width=45,
            image=self._img_load,
            command=lambda: self._on_load_json(),
        )
        self._chk_upload_keys = Checkbutton(
            self,
            text=LBLSPTNUPLOAD,
            variable=self._upload_keys,
        )
        self._chk_send_config = Checkbutton(
            self,
            text=LBLSPTNFP,
            variable=self._send_f9p_config,
        )
        self._chk_disable_nmea = Checkbutton(
            self, text=LBLDISNMEA, variable=self._disable_nmea
        )
        self._lbl_send_command = Label(self, image=self._img_blank)
        self._btn_send_command = Button(
            self,
            image=self._img_send,
            width=45,
            fg=OKCOL,
            command=self._on_send_gnss_config,
            font=self.__app.font_md,
        )

    def _do_layout(self):
        """
        Position widgets in frame.
        """

        self._lbl_gnss.grid(column=0, row=0, columnspan=4, padx=3, pady=2, sticky=W)
        self._lbl_loadjson.grid(column=0, row=1, padx=3, pady=2, sticky=W)
        self._btn_loadjson.grid(column=1, row=1, columnspan=3, padx=3, pady=2, sticky=W)
        ttk.Separator(self).grid(
            column=0, row=2, columnspan=4, padx=2, pady=3, sticky=(W, E)
        )
        self._lbl_curr.grid(column=0, row=3, columnspan=4, padx=3, pady=2, sticky=W)
        self._lbl_key1.grid(column=0, row=4, padx=3, pady=2, sticky=W)
        self._ent_key1.grid(column=1, row=4, columnspan=3, padx=3, pady=2, sticky=W)
        self._lbl_valdate1.grid(column=0, row=5, padx=3, pady=2, sticky=W)
        self._ent_valdate1.grid(column=1, row=5, columnspan=3, padx=3, pady=2, sticky=W)
        ttk.Separator(self).grid(
            column=0, row=6, columnspan=4, padx=2, pady=3, sticky=(W, E)
        )
        self._lbl_next.grid(column=0, row=7, columnspan=3, padx=3, pady=2, sticky=W)
        self._lbl_key2.grid(column=0, row=8, padx=3, pady=2, sticky=W)
        self._ent_key2.grid(column=1, row=8, columnspan=3, padx=2, pady=2, sticky=W)
        self._lbl_valdate2.grid(column=0, row=9, padx=3, pady=2, sticky=W)
        self._ent_valdate2.grid(column=1, row=9, columnspan=3, padx=3, pady=2, sticky=W)
        ttk.Separator(self).grid(
            column=0, row=10, columnspan=4, padx=2, pady=3, sticky=(W, E)
        )
        self._rad_source0.grid(column=0, row=11, padx=3, pady=2, sticky=W)
        self._rad_source1.grid(column=1, row=11, columnspan=3, padx=3, pady=2, sticky=W)
        self._lbl_dgnssto.grid(column=0, row=12, columnspan=3, padx=3, pady=2, sticky=W)
        self._spn_dgnssto.grid(column=1, row=12, columnspan=3, padx=3, pady=2, sticky=W)
        self._chk_upload_keys.grid(column=0, row=13, padx=3, pady=2, sticky=W)
        self._chk_send_config.grid(column=1, row=13, padx=3, pady=2, sticky=W)
        self._chk_disable_nmea.grid(
            column=2, row=13, columnspan=2, padx=3, pady=2, sticky=W
        )
        ttk.Separator(self).grid(
            column=0, row=14, columnspan=4, padx=2, pady=3, sticky=(W, E)
        )
        self._btn_send_command.grid(column=2, row=15, padx=3, pady=2, sticky=W)
        self._lbl_send_command.grid(column=3, row=15, padx=3, pady=2, sticky=W)

    def _reset(self):
        """
        Reset configuration widgets.
        """

        curr = datetime.now()
        self._set_date(self._spartn_valdate1, curr)
        self._set_date(self._spartn_valdate2, curr + timedelta(weeks=4))
        self._spartn_source.set(SPARTN_SOURCE_IP)
        self._upload_keys.set(1)
        self._send_f9p_config.set(1)
        self._disable_nmea.set(0)

        self._poll_config()

    def _attach_events(self):
        """
        Set up event listeners.
        """

        self._spartn_key1.trace_add("write", self._on_update_config)

    def _on_update_config(self, var, index, mode):  # pylint: disable=unused-argument
        """
        Update in-memory configuration if setting is changed.
        """

        self.update()
        self.__app.configuration.set("spartnkey_s", self._spartn_key1.get())
        self.__app.configuration.set(
            "spartnbasedate_n", date2timetag(datetime.now(timezone.utc))
        )

    def _valid_gnss_settings(self) -> bool:
        """
        Validate GNSS receiver settings.

        :return: valid True/False
        :rtype: bool
        """

        valid = True
        # key must be valid hexadecimal and 16 bytes (32 chars) in length
        valid = (
            valid
            & valid_entry(self._ent_key1, VALHEX)
            & valid_entry(self._ent_key1, VALLEN, SPARTN_KEYLEN * 2, SPARTN_KEYLEN * 2)
        )
        # from date must be valid date in format YYYYMMDD
        valid = valid & valid_entry(self._ent_valdate1, VALDMY)
        valid = (
            valid
            & valid_entry(self._ent_key2, VALHEX)
            & valid_entry(self._ent_key2, VALLEN, SPARTN_KEYLEN * 2, SPARTN_KEYLEN * 2)
        )
        valid = valid & valid_entry(self._ent_valdate2, VALDMY)

        if not valid:
            self.__container.set_status("ERROR - invalid settings", ERRCOL)

        return valid

    def _format_cfggnss(self) -> UBXMessage:
        """
        Format UBX CFG-VALSET message to configure GNSS receiver.
        """

        cfgdata = []
        cfgdata.append(("CFG_NAVHPG_DGNSSMODE", RTK_FIXED))
        # select SPARTN source (L-band or IP)
        cfgdata.append(("CFG_SPARTN_USE_SOURCE", self._spartn_source.get()))
        # set DGNSS Timeout (default = 60, higher can be better with L-Band)
        cfgdata.append(("CFG_NAVSPG_CONSTR_DGNSSTO", self._dgnssto.get()))
        # enable SPARTN and UBX on selected ports
        for port in self._ports:
            cfgdata.append((f"CFG_{port}INPROT_UBX", 1))
            cfgdata.append((f"CFG_{port}INPROT_RTCM3X", 1))
            cfgdata.append((f"CFG_{port}INPROT_SPARTN", 1))
            cfgdata.append((f"CFG_MSGOUT_UBX_RXM_COR_{port}", 1))

            # if NMEA disabled, remove it from out ports and enable
            # a couple of UBX NAV messages instead
            if self._disable_nmea.get():
                cfgdata.append((f"CFG_{port}OUTPROT_NMEA", 0))
                cfgdata.append((f"CFG_MSGOUT_UBX_NAV_PVT_{port}", 1))
                cfgdata.append((f"CFG_MSGOUT_UBX_NAV_DOP_{port}", 1))
                cfgdata.append((f"CFG_MSGOUT_UBX_NAV_SAT_{port}", 4))
            else:
                cfgdata.append((f"CFG_{port}OUTPROT_NMEA", 1))
                # cfgdata.append((f"CFG_MSGOUT_UBX_NAV_PVT_{port}", 0))
                # cfgdata.append((f"CFG_MSGOUT_UBX_NAV_DOP_{port}", 0))
                # cfgdata.append((f"CFG_MSGOUT_UBX_NAV_SAT_{port}", 0))

        msg = UBXMessage.config_set(SET_LAYER_RAM, TXN_NONE, cfgdata)
        return msg

    def _format_rxmspartn(self) -> UBXMessage:
        """
        Format UBX RXM-SPARTN-KEY message.
        """

        version = val2bytes(1, U1)
        numKeys = val2bytes(2, U1)
        key1 = bytes.fromhex(self._spartn_key1.get())
        keylen1 = val2bytes(len(key1), U1)
        wno1, tow1 = date2wnotow(self._get_date(self._spartn_valdate1))
        wno1b = val2bytes(wno1, U2)
        tow1b = val2bytes(tow1, U4)
        key2 = bytes.fromhex(self._spartn_key2.get())
        keylen2 = val2bytes(len(key2), U1)
        wno2, tow2 = date2wnotow(self._get_date(self._spartn_valdate2))
        wno2b = val2bytes(wno2, U2)
        tow2b = val2bytes(tow2, U4)

        payload = (
            version
            + numKeys
            + RESERVED0
            + RESERVED1
            + keylen1
            + wno1b
            + tow1b
            + RESERVED1
            + keylen2
            + wno2b
            + tow2b
            + key1
            + key2
        )
        msg = UBXMessage("RXM", RXMMSG, SET, payload=payload)
        return msg

    def _on_send_gnss_config(self):
        """
        Send config to GNSS receiver (e.g. F9P).
        """

        if not self._valid_gnss_settings():
            return

        if self.__app.conn_status != CONNECTED:
            self.__container.set_status(NOTCONN, ERRCOL)
            # return

        if self._send_f9p_config.get():
            msgc = "config"
            msg = self._format_cfggnss()
            self._send_gnss_command(msg)
        else:
            msgc = ""
        if self._upload_keys.get():
            msgk = "keys"
            msg = self._format_rxmspartn()
            self._send_gnss_command(msg)
            # poll RXM-SPARTN-KEY to check keys have loaded
            msg = UBXMessage("RXM", RXMMSG, POLL)
            self._send_gnss_command(msg)
        else:
            msgk = ""
        if msgk == "" and msgc == "":
            self.__container.set_status(NULLSEND, ERRCOL)
            return
        if msgk != "" and msgc != "":
            msga = " and "
        else:
            msga = ""
        self.__container.set_status(f"{(msgk + msga + msgc).capitalize()} sent", OKCOL)
        self._lbl_send_command.config(image=self._img_pending)
        for msgid in ("RXM-SPARTNKEY", "CFG-VALGET", "ACK-ACK", "ACK-NAK"):
            self.__container.set_pending(msgid, SPARTN_GNSS)

    def _send_gnss_command(self, msg: UBXMessage):
        """
        Send command to GNSS receiver.
        """

        self.__app.gnss_outqueue.put(msg.serialize())

    def _set_date(self, val: StringVar, dat: datetime):
        """
        Set date string in entry field.
        """

        val.set(dat.strftime("%Y%m%d"))

    def _get_date(self, val: StringVar) -> datetime:
        """
        Get datetime from entry field.
        """

        dat = val.get()
        return datetime(int(dat[0:4]), int(dat[4:6]), int(dat[6:8]))

    def _poll_config(self):
        """
        Poll receiver for current RXM-SPARTN-KEY configuration.
        """

        msg = UBXMessage("RXM", RXMMSG, POLL)
        self.__app.gnss_outqueue.put(msg.serialize())
        self.__container.set_pending(RXMMSG, SPARTN_GNSS)

    def update_status(self, msg: UBXMessage):
        """
        Update pending confirmation status.
        :param UBXMessage msg: UBX config message
        """

        if msg.identity == RXMMSG:
            self._lbl_send_command.config(image=self._img_confirmed)
            if msg.numKeys == 2:
                keydata = parse_rxmspartnkey(msg)  # key1, date1, key2, date2
                self._spartn_key1.set(keydata[0][0])
                self._spartn_valdate1.set(keydata[0][1].strftime("%Y%m%d"))
                self._spartn_key2.set(keydata[1][0])
                self._spartn_valdate2.set(keydata[1][1].strftime("%Y%m%d"))
                # save latest key in configuration settings
                self.__app.configuration.set("spartnkey_s", self._spartn_key1.get())
                col = OKCOL
            else:
                col = ERRCOL
            self.__container.set_status(CONFIGRXM.format(RXMMSG, msg.numKeys), col)
        elif msg.identity == "ACK-ACK":
            self._lbl_send_command.config(image=self._img_confirmed)
            self.__container.set_status(CONFIGOK.format(CFGSET), OKCOL)
        elif msg.identity == "ACK-NAK":
            self._lbl_send_command.config(image=self._img_warn)
            self.__container.set_status(CONFIGBAD.format(CFGSET), ERRCOL)
        self.update_idletasks()

    def _on_load_json(self):
        """
        Load SPARTN decryption keys from JSON file.
        """
        # pylint: disable=unused-variable

        jsonfile = self.__app.file_handler.open_file(
            "spartnjson", (("json files", "*.json"), ("all files", "*.*"))
        )
        if jsonfile is None:
            return

        try:
            spc = SpartnJsonConfig(jsonfile)
            # strip scheme and port from server name
            scheme, host, port = spc.server.split(":")
            self.__container.server = host.replace("//", "")
            self.__container.clientid = spc.clientid
            (key, start, _) = spc.current_key
            self._spartn_key1.set(key)
            self._spartn_valdate1.set(start.strftime("%Y%m%d"))
            (key, start, _) = spc.next_key
            self._spartn_key2.set(key)
            self._spartn_valdate2.set(start.strftime("%Y%m%d"))
            self.__container.set_status(DLGJSONOK.format(jsonfile), OKCOL)
        except Exception as err:  # pylint: disable=broad-exception-caught
            self.__container.set_status(DLGJSONERR.format(err), ERRCOL)
